"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for module, drain_swamp.backend_setupttools

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_backend_abc.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_backend_abc.py

"""

import logging
import logging.config
import sys
from pathlib import (
    Path,
    PurePath,
)
from unittest.mock import MagicMock

import pytest

from drain_swamp._run_cmd import run_cmd
from drain_swamp.backend_abc import (
    BackendType,
    ensure_folder,
    folders_additional_init,
    folders_implied_init,
    get_optionals_cli,
    get_optionals_pyproject_toml,
    get_required_cli,
    get_required_pyproject_toml,
    try_dict_update,
)
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.exceptions import (
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)
from drain_swamp.parser_in import TomlParser
from drain_swamp.snip import ReplaceResult

if sys.version_info >= (3, 9):  # pragma: no cover
    from collections.abc import Sequence
else:  # pragma: no cover
    from typing import Sequence

testdata_fix_suffix = [
    ("md", ".md"),
    (".md", ".md"),
]
ids_fix_suffix = [
    "without prefix period",
    "with prefix period",
]


@pytest.mark.parametrize(
    "suffix, expected",
    testdata_fix_suffix,
    ids=ids_fix_suffix,
)
def test_fix_suffix(suffix, expected):
    """Does not barf given multiple suffixes e.g. .tar.gz"""
    # pytest --showlocals --log-level INFO -k "test_fix_suffix" tests
    actual = BackendType.fix_suffix(suffix)
    assert actual == expected


testdata_determine_backend = [
    (
        Path(__file__).parent.joinpath("_good_files", "backend-only.pyproject_toml"),
        "setuptools",
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "setuptools",
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "requires-none.pyproject_toml"),
        "setuptools",
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "nonsense-keys.pyproject_toml"),
        "setuptools",
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files", "thin-wrap-backend.pyproject_toml"
        ),
        "setuptools",
    ),
]
ids_determine_backend = [
    f"""{t_backend[0].name.rsplit(".", 1)[0]}{t_backend[1]}"""
    for t_backend in testdata_determine_backend
]


@pytest.mark.parametrize(
    "path, expected",
    testdata_determine_backend,
    ids=ids_determine_backend,
)
def test_determine_backend(path, expected):
    # pytest --showlocals --log-level INFO -k "test_determine_backend" tests
    tp = TomlParser(path)
    d_pyproject_toml = tp.d_pyproject_toml
    actual = BackendType.determine_backend(d_pyproject_toml)
    assert actual == expected


def test_try_dict_update(tmp_path, caplog, has_logging_occurred, prepare_folders_files):
    # pytest --showlocals --log-level INFO -k "test_try_dict_update" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # is_bypass == True; skips validating exists and is_file
    target_x = "prod"
    relative_x_path = "requirements/prod.in"
    dict_cli = dict()
    try_dict_update(dict_cli, tmp_path, target_x, Path(relative_x_path), is_bypass=True)
    assert len(dict_cli.keys()) == 1
    assert dict_cli[target_x] == tmp_path / Path(relative_x_path)

    # is_bypass None of unsupported type --> False
    prepare_folders_files((relative_x_path,), tmp_path)

    no_cheatings = (
        None,
        1.12345,
        False,
    )
    for is_bypass in no_cheatings:
        dict_cli_actual = dict()
        try_dict_update(
            dict_cli_actual,
            tmp_path,
            target_x,
            Path(relative_x_path),
            is_bypass=is_bypass,
        )
        # assert has_logging_occurred(caplog)
        assert len(dict_cli_actual.keys()) == 1
        assert dict_cli_actual[target_x] == tmp_path / Path(relative_x_path)


testdata_get_required_pyproject_toml = [
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        ("requirements/prod.in",),
        ("prod", Path("requirements/prod.in")),
    ),
]
ids_get_required_pyproject_toml = ("is_bypass True and False",)


@pytest.mark.parametrize(
    "path, lst_create_it, t_expected",
    testdata_get_required_pyproject_toml,
    ids=ids_get_required_pyproject_toml,
)
def test_get_required_pyproject_toml(
    path,
    lst_create_it,
    t_expected,
    tmp_path,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_get_required_pyproject_toml" -v tests
    tp = TomlParser(path)
    d_pyproject_toml = tp.d_pyproject_toml
    # is_bypass == True; skips validating exists and is_file
    # tuple[str, Path] | None
    t_actual = get_required_pyproject_toml(
        d_pyproject_toml,
        tmp_path,
        is_bypass=True,
    )
    if t_actual is None:
        assert t_actual is None
    else:
        assert isinstance(t_actual, Sequence)
        assert t_actual[0] == t_expected[0]
        assert t_actual[1] == tmp_path / t_expected[1]

    # prepare
    prepare_folders_files(lst_create_it, tmp_path)

    # is_bypass == False
    no_cheatings = (
        None,
        1.12345,
        False,
    )
    for is_bypass in no_cheatings:
        t_actual = get_required_pyproject_toml(
            d_pyproject_toml,
            tmp_path,
            is_bypass=is_bypass,
        )

        if t_actual is None:
            assert t_actual is None
        else:
            assert isinstance(t_actual, Sequence)
            assert t_actual[0] == t_expected[0]
            assert t_actual[1] == tmp_path / t_expected[1]


testdata_get_required_cli = [
    (
        ("requirements/prod.in",),
        ("prod", Path("requirements/prod.in")),
        ("prod", Path("requirements/prod.in")),
    ),
    (
        ("requirements/prod.in",),
        ("prod", "requirements/prod.in"),
        ("prod", Path("requirements/prod.in")),
    ),
    (
        ("requirements/prod.in",),
        ("prod", 1.2345),
        None,
    ),
    (
        ("requirements/prod.in",),
        None,
        None,
    ),
]
ids_get_required_cli = (
    "relative_path is Path",
    "relative_path is str",
    "relative_path is float",
    "None rather than Sequence",
)


@pytest.mark.parametrize(
    "lst_create_it, t_required, t_expected",
    testdata_get_required_cli,
    ids=ids_get_required_cli,
)
def test_get_required_cli(
    lst_create_it,
    t_required,
    t_expected,
    tmp_path,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_get_required_cli" -v tests
    # bypass True
    t_actual = get_required_cli(tmp_path, required=t_required, is_bypass=True)
    if t_actual is None:
        assert t_actual is None
    else:
        assert t_actual[0] == t_expected[0]
        assert t_actual[1] == tmp_path / t_expected[1]

    # bypass False
    #    prepare
    prepare_folders_files(lst_create_it, tmp_path)

    no_cheatings = (
        None,
        1.12345,
        False,
    )
    for is_bypass in no_cheatings:
        t_actual = get_required_cli(tmp_path, required=t_required, is_bypass=is_bypass)
        if t_actual is None:
            assert t_actual is None
        else:
            assert isinstance(t_actual, Sequence)
            assert t_actual[0] == t_expected[0]
            assert t_actual[1] == tmp_path / t_expected[1]


testdata_required = [
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        None,
        ("requirements/prod.in",),
        ("prod", Path("requirements/prod.in")),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "",
        ("requirements/prod.in",),
        ("prod", Path("requirements/prod.in")),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        "    ",
        ("requirements/prod.in",),
        ("prod", Path("requirements/prod.in")),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        1.12345,
        ("requirements/prod.in",),
        ("prod", Path("requirements/prod.in")),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        ("alt", "requirements/alt.in"),
        ("requirements/prod.in", "requirements/alt.in"),
        ("alt", Path("requirements/alt.in")),
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        ("alt", "requirements/alt.in"),
        [],
        None,
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        ("alt", Path("requirements/alt.in")),
        [],
        None,
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        ("alt", Path("requirements/alt.in")),
        ("requirements/prod.in", "requirements/alt.in"),
        ("alt", Path("requirements/alt.in")),
    ),
]
ids_required = (
    "cli None",
    "cli empty string",
    "cli excessive whitespace",
    "cli unsupported type",
    "prioritize cli over pyproject.toml",
    "no required file exists",
    "cli relative path is Path; no actual file",
    "cli relative path is Path; actual file",
)


@pytest.mark.parametrize(
    "path, cli_required, lst_create_it, t_expected",
    testdata_required,
    ids=ids_required,
)
def test_get_required(
    path,
    cli_required,
    lst_create_it,
    t_expected,
    tmp_path,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_get_required" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare
    prepare_folders_files(lst_create_it, tmp_path)

    tp = TomlParser(path)
    d_pyproject_toml = tp.d_pyproject_toml
    t_actual = BackendType.get_required(
        d_pyproject_toml, tmp_path, required=cli_required
    )
    if t_actual is None:
        # assert has_logging_occurred(caplog)
        assert t_actual == t_expected
    else:
        target_a, abspath_a = t_actual
        relpath_a = abspath_a.relative_to(tmp_path)
        # assert has_logging_occurred(caplog)
        assert (target_a, relpath_a) == t_expected


testdata_get_optionals_cli = (
    (
        {},
        {},
    ),
    (
        {"tox": Path("requirements/tox.in"), "kit": Path("requirements/kit.in")},
        {"tox": Path("requirements/tox.in"), "kit": Path("requirements/kit.in")},
    ),
    (
        {"tox": "requirements/tox.in", "kit": "requirements/kit.in"},
        {"tox": Path("requirements/tox.in"), "kit": Path("requirements/kit.in")},
    ),
    (
        {"tox": None, "kit": 1.1234, "prod": 1},
        {},
    ),
    (
        {"tox": None, "kit": None},
        {},
    ),
)
ids_get_optionals_cli = (
    "empty dict",
    "values relative Path",
    "values relative path str",
    "values unsupported types",
    "values None",
)


@pytest.mark.parametrize(
    "d_pairs, d_expected",
    testdata_get_optionals_cli,
    ids=ids_get_optionals_cli,
)
def test_get_optionals_cli(d_pairs, d_expected, tmp_path, prepare_folders_files):
    # pytest --showlocals --log-level INFO -k "test_get_optionals_cli" -v tests

    # seq_rel_paths not a Sequence
    invalids = (
        {"tox": None},
        {"kit": 1.2345},
        {"prod": 1},
    )
    d_invalids_expected = dict()
    for d_invalid in invalids:
        get_optionals_cli(
            d_invalids_expected,
            tmp_path,
            d_invalid,
        )
        assert len(d_invalids_expected.keys()) == 0

    # prepare
    if isinstance(d_pairs, dict):
        #    dict_values --> list
        seq_prepare_these = list(d_pairs.values())
        prepare_folders_files(seq_prepare_these, tmp_path)

    # act
    d_cli_actual = dict()

    get_optionals_cli(
        d_cli_actual,
        tmp_path,
        d_pairs,
    )

    # d_expected_abs_paths -- target/path_rel --> target / abs path
    # d_cli_actual         -- target / abs path
    d_expected_abs_paths = dict()
    for target, path_rel in d_expected.items():
        abs_path = tmp_path / path_rel
        d_expected_abs_paths.update({target: abs_path})

    assert d_cli_actual == d_expected_abs_paths


testdata_get_optionals_pyproject_toml = (
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        5,
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "nonsense-keys.pyproject_toml"),
        5,
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "requires-none.pyproject_toml"),
        5,
    ),
)
ids_get_optionals_pyproject_toml = [
    f"""{t_backend[0].name.rsplit(".", 1)[0]}{t_backend[1]}"""
    for t_backend in testdata_get_optionals_pyproject_toml
]


@pytest.mark.parametrize(
    "path, expected_count",
    testdata_get_optionals_pyproject_toml,
    ids=ids_get_optionals_pyproject_toml,
)
def test_get_optionals_pyproject_toml(
    path,
    expected_count,
    tmp_path,
    caplog,
    has_logging_occurred,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_get_optionals_pyproject_toml" -v tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    tp = TomlParser(path)
    d_pyproject_toml = tp.d_pyproject_toml

    # is_bypass True
    d_actual = dict()
    get_optionals_pyproject_toml(
        d_actual,
        d_pyproject_toml,
        tmp_path,
        is_bypass=True,
    )
    # assert has_logging_occurred(caplog)
    assert len(d_actual.keys()) == expected_count

    """prepare

    seq_expected contains rel path need abs paths. Create folders and files
    """
    d_prepare_these = {
        "pip": "requirements/pip.in",
        "pip_tools": "requirements/pip-tools.in",
        "dev": "requirements/dev.in",
        "manage": "requirements/manage.in",
        "docs": "docs/requirements.in",
    }
    seq_prepare_these = list(d_prepare_these.values())
    prepare_folders_files(seq_prepare_these, tmp_path)

    d_expected_abspaths = dict()

    # is_bypass unsupported type --> False
    is_bypasses = (
        None,
        1.2345,
        1,
        False,
    )
    d_expected_abspaths = dict()
    for target, rel_path in d_prepare_these.items():
        path_abs = tmp_path / rel_path
        d_expected_abspaths.update({target: path_abs})

    for is_bypass in is_bypasses:
        d_actual = dict()
        get_optionals_pyproject_toml(
            d_actual,
            d_pyproject_toml,
            tmp_path,
            is_bypass=is_bypass,
        )
        # assert has_logging_occurred(caplog)

        d_actual == d_expected_abspaths


testdata_get_optional = [
    (
        Path(__file__).parent.joinpath("_good_files", "complete.pyproject_toml"),
        {"tox": Path("requirements/tox.in"), "kit": Path("requirements/kit.in")},
    ),
    (
        Path(__file__).parent.joinpath("_good_files", "nonsense-keys.pyproject_toml"),
        {"tox": Path("requirements/tox.in"), "kit": Path("requirements/kit.in")},
    ),
]
ids_get_optional = (
    "Path cli supplied tox and kit",
    "pyproject.toml nonsense-keys",
)


@pytest.mark.parametrize(
    "path, d_cli",
    testdata_get_optional,
    ids=ids_get_optional,
)
def test_get_optionals_both(
    path,
    d_cli,
    tmp_path,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_get_optionals" -v tests
    tp = TomlParser(path)
    d_pyproject_toml = tp.d_pyproject_toml

    # Prepare -- cli
    seq_prepare_these = list(d_cli.values())
    prepare_folders_files(seq_prepare_these, tmp_path)

    # Prepare -- pyproject.toml
    d_pypro = dict()
    get_optionals_pyproject_toml(
        d_pypro,
        d_pyproject_toml,
        tmp_path,
        is_bypass=True,
    )
    seq_prepare_these = list(d_pypro.values())
    prepare_folders_files(seq_prepare_these, tmp_path)

    # cli suppliments pyproject.toml
    for target, rel_path in d_cli.items():
        is_pathlike = isinstance(rel_path, str) or issubclass(type(rel_path), PurePath)
        if is_pathlike:
            d_pypro[target] = tmp_path.joinpath(rel_path)

    d_actual = BackendType.get_optionals(
        d_pyproject_toml,
        tmp_path,
        d_cli,
    )
    assert d_actual == d_pypro


testdata_folders_implied_init = (
    (
        {
            "pip": "requirements/pip.in",
            "pip_tools": "requirements/pip-tools.in",
            "dev": "requirements/dev.in",
            "manage": "requirements/manage.in",
            "docs": "docs/requirements.in",
        },
        ("prod", "requirements/prod.in"),
        {Path("requirements"), Path("docs")},
    ),
)
ids_folders_implied_init = ("normal optionals and required",)


@pytest.mark.parametrize(
    "d_optionals, t_required, s_expected",
    testdata_folders_implied_init,
    ids=ids_folders_implied_init,
)
def test_folders_implied_init(
    d_optionals,
    t_required,
    s_expected,
    tmp_path,
    prepare_folders_files,
):
    # pytest --showlocals --log-level INFO -k "test_folders_implied_init" -v tests
    # input lacks abs path. Including expected

    # prepare
    seq_prepare_these = list(d_optionals.values())
    prepare_folders_files(seq_prepare_these, tmp_path)
    if t_required is not None:
        seq_prepare_these = (t_required[1],)
    else:
        seq_prepare_these = t_required
    prepare_folders_files(seq_prepare_these, tmp_path)

    # act
    d_optionals_abspath = {}
    for key, value in d_optionals.items():
        d_optionals_abspath[key] = tmp_path / value

    if t_required is not None:
        t_required_abs_path = (t_required[0], tmp_path / t_required[1])
    else:
        t_required_abs_path = None

    s_actual = folders_implied_init(tmp_path, d_optionals_abspath, t_required_abs_path)
    assert s_actual == s_expected


testdata_folders_additional_init = (
    (
        {Path("requirements"), Path("docs")},
        (Path("requirements"), Path("docs"), Path("ci")),
        {Path("ci")},
    ),
    (
        {Path("requirements"), Path("docs")},
        (Path("ci"),),
        {Path("ci")},
    ),
)
ids_folders_additional_init = (
    "Folders contains all folders",
    "Folders contains just the additional folders",
)


@pytest.mark.parametrize(
    "s_folders_implied, t_additional_folders, s_expected",
    testdata_folders_additional_init,
    ids=ids_folders_additional_init,
)
def test_folders_additional_init(
    s_folders_implied,
    t_additional_folders,
    s_expected,
    tmp_path,
):
    # pytest --showlocals --log-level INFO -k "test_folders_additional_init" -v tests
    # prepare
    for path_dir_implied in s_folders_implied:
        abs_path = tmp_path / path_dir_implied
        abs_path.mkdir(parents=True, exist_ok=True)
    #    Actual code is not naive about existance of implied folders
    for path_dir_additional in t_additional_folders:
        abs_path = tmp_path / path_dir_additional
        abs_path.mkdir(parents=True, exist_ok=True)

    # Should contain relative paths. Which is normal situation / usage
    s_actual = folders_additional_init(
        tmp_path,
        s_folders_implied,
        t_additional_folders,
    )
    assert s_actual == s_expected

    # Absolute paths. This is abnormal usage. Academic use case
    lst_additional_folders = []
    if t_additional_folders is not None:
        for additional_folders_relpath in t_additional_folders:
            lst_additional_folders.append(tmp_path / additional_folders_relpath)
        t_additional_folders_abspath = tuple(lst_additional_folders)
    else:
        t_additional_folders_abspath = t_additional_folders

    s_actual = folders_additional_init(
        tmp_path,
        s_folders_implied,
        t_additional_folders_abspath,
    )
    assert s_actual == s_expected


def test_ensure_folder(tmp_path):
    # pytest --showlocals --log-level INFO -k "test_ensure_folder" -v tests
    # folder
    path_actual = ensure_folder(tmp_path)
    assert issubclass(type(path_actual), PurePath)

    # file
    path_file = tmp_path.joinpath("todd.in")
    path_file.touch()
    path_actual = ensure_folder(path_file)
    assert path_actual.is_dir() and path_actual == tmp_path

    # symlink (by someone with a grudge against todd and a fondness for harrassment)
    p = tmp_path.joinpath("todd_wears_drag_and_drinks_bud_lite.in")
    p.symlink_to(path_file)
    with pytest.raises(NotADirectoryError):
        ensure_folder(p)

    # MagicMock with a return_value which isn't a Path
    mock = MagicMock(wraps=ensure_folder)
    with pytest.raises(TypeError):
        mock("Hello World!")

    # unsupported type
    invalids = (
        None,
        1.2345,
        "Hello World!",
    )
    for invalid in invalids:
        with pytest.raises(TypeError):
            ensure_folder(invalid)


testdata_read = (
    pytest.param(
        Path(__file__).parent.joinpath("_bad_files", "backend_only.pyproject_toml"),
        marks=pytest.mark.xfail(raises=PyProjectTOMLParseError),
    ),
    pytest.param(
        Path(__file__).parent.joinpath("_good_files", "nonexistent.pyproject_toml"),
        marks=pytest.mark.xfail(raises=PyProjectTOMLReadError),
    ),
)
ids_read = (
    "unparsable pyproject-toml",
    "no such pyproject-toml",
)


@pytest.mark.parametrize(
    "path_toml_src",
    testdata_read,
    ids=ids_read,
)
def test_read(path_toml_src, tmp_path, prep_pyproject_toml):
    # pytest --showlocals --log-level INFO -k "test_read" -v tests
    if not path_toml_src.exists():
        path_f = tmp_path
    else:
        path_f = prep_pyproject_toml(path_toml_src, tmp_path)
    BackendType.read(path_f)


testdata_is_locked = (
    (
        "no_snippet_with_id_none.txt",
        None,
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there george\n# @@@ end\nzzzzzzz\n",
        ReplaceResult.NO_MATCH,
    ),
    (
        "no_snippet_with_that_id.txt",
        "ted",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there george\n# @@@ end\nzzzzzzz\n",
        ReplaceResult.NO_MATCH,
    ),
    (
        "snippet_with_id_all_unlocked.txt",
        "ted",
        """zzzzzzzzzzzzz\n# @@@ editable ted\ndependencies = { file = ["requirements/prod.unlock"] }
optional-dependencies.pip = { file = ["requirements/pip.unlock"] }
optional-dependencies.pip_tools = { file = ["requirements/pip-tools.unlock"] }
optional-dependencies.dev = { file = ["requirements/dev.unlock"] }
optional-dependencies.manage = { file = ["requirements/manage.unlock"] }
optional-dependencies.docs = { file = ["docs/requirements.unlock"] }\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there george\n# @@@ end\nzzzzzzz\n""",
        False,
    ),
    (
        "snippet_with_id_all_locked.txt",
        "ted",
        """zzzzzzzzzzzzz\n# @@@ editable ted\ndependencies = { file = ["requirements/prod.lock"] }
optional-dependencies.pip = { file = ["requirements/pip.lock"] }
optional-dependencies.pip_tools = { file = ["requirements/pip-tools.lock"] }
optional-dependencies.dev = { file = ["requirements/dev.lock"] }
optional-dependencies.manage = { file = ["requirements/manage.lock"] }
optional-dependencies.docs = { file = ["docs/requirements.lock"] }\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there george\n# @@@ end\nzzzzzzz\n""",
        True,
    ),
    (
        "snippet_with_id_unlocked_no_matches.txt",
        "ted",
        """zzzzzzzzzzzzz\n# @@@ editable ted\ndependencies = { file = ["requirements/prod.pip"] }
optional-dependencies.pip = { file = ["requirements/pip.pip"] }
optional-dependencies.pip_tools = { file = ["requirements/pip-tools.pip"] }
optional-dependencies.dev = { file = ["requirements/dev.pip"] }
optional-dependencies.manage = { file = ["requirements/manage.pip"] }
optional-dependencies.docs = { file = ["docs/requirements.pip"] }\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there george\n# @@@ end\nzzzzzzz\n""",
        False,
    ),
)
testdata_is_locked2 = (
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete.pyproject_toml",
        ),
        True,
    ),
    (
        Path(__file__).parent.joinpath(
            "_good_files",
            "complete-manage-pip-prod-unlock.pyproject_toml",
        ),
        False,
    ),
    pytest.param(
        Path(__file__).parent.joinpath(
            "_bad_files",
            "static_dependencies.pyproject_toml",
        ),
        None,
        marks=pytest.mark.xfail(raises=AssertionError),
    ),
)
ids_is_locked2 = (
    "Locked",
    "Unlocked",
    "static dependencies. no tool.setuptools.dynamic section",
)


@pytest.mark.parametrize(
    "path_config, expected",
    testdata_is_locked2,
    ids=ids_is_locked2,
)
def test_is_locked(path_config, expected, tmp_path, prep_pyproject_toml):
    # pytest --showlocals --log-level INFO -k "test_is_locked" -v tests
    # path_config invalids. Must be: pathlib.Path, absolute, to a file
    invalids = (
        1.1234,
        1,
        None,
    )
    for invalid in invalids:
        with pytest.raises(PyProjectTOMLReadError):
            BackendType.is_locked(invalid)

    # is_file_ok fails --> PyProjectTOMLReadError
    with pytest.raises(PyProjectTOMLReadError):
        BackendType.is_locked(tmp_path)

    # prepare
    #    does not check existance of requirements files
    path_config_in_tmp = prep_pyproject_toml(path_config, tmp_path)

    actual = BackendType.is_locked(path_config_in_tmp)
    assert actual is expected


def test_resolve_symlinks(tmp_path, prep_pyproject_toml, prepare_folders_files):
    # pytest --showlocals --log-level INFO -k "test_resolve_symlinks" -v tests
    path_config_src = Path(__file__).parent.joinpath(
        "_good_files",
        "complete-lnk-files.pyproject_toml",
    )
    path_f = prep_pyproject_toml(path_config_src, tmp_path)

    d_ins = {
        "prod": "requirements/prod.in",
        "pip": "requirements/pip.in",
        "tox": "requirements/tox.in",
        "manage": "requirements/manage.in",
    }
    seq_ins = list(d_ins.values())
    prepare_folders_files(seq_ins, tmp_path)

    seq_unlocks = (
        "requirements/prod.unlock",
        "requirements/pip.unlock",
        "requirements/tox.unlock",
        "requirements/manage.unlock",
        "requirements/prod.lock",
        "requirements/pip.lock",
        "requirements/tox.lock",
        "requirements/manage.lock",
    )
    prepare_folders_files(seq_unlocks, tmp_path)

    # order matters
    cmds = (
        (
            ("pipenv-unlock", "refresh", "--path", str(tmp_path), "--set-lock", "0"),
            False,
        ),
        (
            ("pipenv-unlock", "refresh", "--path", str(tmp_path), "--set-lock", "1"),
            True,
        ),
    )
    for cmd, expected in cmds:
        run_cmd(cmd, cwd=tmp_path)
        actual = BackendType.is_locked(path_f)
        assert actual is expected



import abc
import logging
import sys
from collections.abc import (
    Generator,
    Iterator,
)
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Final,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

__all__ = ("BackendType",)
_logger: Final[logging.Logger]
is_module_debug: Final[bool]
entrypoint_name: Final[str]

def try_dict_update(
    set_both: dict[str, Path],
    path_config: Path,
    target_x: str,
    path_relative_x: Path,
    is_bypass: Any | None = False,
) -> None: ...
def get_optionals_cli(
    d_both: dict[str, Path],
    path_config: Path,
    optionals: dict[str, Any],
) -> None: ...
def get_optionals_pyproject_toml(
    d_both: dict[str, Path],
    d_pyproject_toml: dict[str, Any],
    path_config: Path,
    is_bypass: bool | None = False,
) -> None: ...
def get_required_pyproject_toml(
    d_pyproject_toml: dict[str, Any],
    path_config: Path,
    is_bypass: bool | None = False,
) -> tuple[str, Path] | None: ...
def get_required_cli(
    path_config: Path,
    required: tuple[str, Any] | None = None,
    is_bypass: bool | None = False,
) -> tuple[str, Path] | None: ...
def folders_implied_init(
    parent_dir: Path,
    optionals: dict[str, Path],
    required: tuple[str, Path] | None = None,
) -> set[Path]: ...
def folders_additional_init(
    parent_dir: Path,
    folders_implied: set[Path],
    additional_folders: tuple[Path] = ...,
) -> set[Path]: ...
def ensure_folder(val: Any) -> Path: ...

class BackendType(abc.ABC):
    _path_required: tuple[str, Path] | None
    _paths_optional: dict[str, Path]
    _path_config: Path
    _parent_dir: Path
    _folders_implied: set[Path]
    _folders_additional: set[Path]
    BACKEND_NAME: ClassVar[str]
    PYPROJECT_TOML_SECTION_NAME: ClassVar[str]

    @staticmethod
    def read(
        path_config: Path,
    ) -> tuple[dict[str, Any], Path]: ...
    @staticmethod
    def load_factory(
        path_config: Path,
        required: tuple[str, Path] | None = None,
        optionals: dict[str, Path] = {},
        parent_dir: Path | None = None,
        additional_folders: tuple[Path, ...] = (),
    ) -> BackendType: ...
    def load(
        self,
        d_pyproject_toml: dict[str, Any],
        required: tuple[str, Path] | None = None,
        optionals: dict[str, Path] = {},
        parent_dir: Path | None = None,
        additional_folders: tuple[Path, ...] = (),
    ) -> None: ...
    @staticmethod
    def determine_backend(d_pyproject_toml: dict[str, Any]) -> str: ...
    @staticmethod
    def get_required(
        d_pyproject_toml: dict[str, Any],
        path_config: Path,
        required: tuple[str, Any] | None = None,
    ) -> tuple[str, Path] | None: ...
    @staticmethod
    def get_optionals(
        d_pyproject_toml: dict[str, Any],
        path_config: Path,
        optionals: dict[str, Any],
    ) -> dict[str, Path]: ...
    @property
    @abc.abstractmethod
    def backend(self) -> str: ...
    @property
    def path_config(self) -> Path: ...
    @property
    def parent_dir(self) -> Path: ...
    @parent_dir.setter
    def parent_dir(
        self,
        parent_dir: Path | None = None,
    ) -> None: ...
    @property
    def required(self) -> tuple[str, Path] | None: ...
    @property
    def optionals(self) -> dict[str, Path]: ...
    @property
    def folders_implied(self) -> set[Path]: ...
    @property
    def folders_additional(self) -> set[Path]: ...
    @classmethod
    def __subclasshook__(cls, C: Any) -> bool: ...
    @classmethod
    def get_registered(cls) -> Iterator[type[Self]]: ...
    @staticmethod
    def fix_suffix(suffix: str) -> str: ...
    def in_files(self) -> Generator[Path, None, None]: ...
    @staticmethod
    def is_locked(
        path_config: Path,
    ) -> bool: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

``pyproject.toml``

.. code-block:: text

   [project.entry-points."setuptools.finalize_distribution_options"]
   drain_swamp = "drain_swamp.monkey.wrap_infer_version:infer_version"

infer_version sets the dist.metadata.version. Would like to do more ...

.. seealso::

   `[alter Distribution object] <https://setuptools.pypa.io/en/latest/userguide/extension.html#customizing-distribution-options>`_

   Command and sub-command classes
   https://setuptools.pypa.io/en/latest/userguide/extension.html#customizing-commands

.. code-block:: text

   setuptools_scm._integration.setuptools.infer_version
     _config.Configuration.from_file
     _assign_version

.. seealso::

   https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_integration/setuptools.py
   https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_integration/pyproject_reading.py
   https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_config.py
   https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_get_version_impl.py
   https://github.com/pypa/setuptools/blob/e9f0be98ea4faaba4a7b2d07ba994a81fde8f42f/setuptools/build_meta.py#L161

   https://github.com/pypa/setuptools/issues/2491
   https://github.com/pypa/setuptools/discussions/4083
   https://github.com/pypa/setuptools/issues/3896#issuecomment-1656714771

   SOLUTIONS

   - pass config_settings through backend
     https://github.com/pypa/setuptools/commit/fc95b3b83d6d5b561dc0a356995edf4c99785a6f

   - throw away setup.cfg file
     https://github.com/pypa/setuptools/issues/3896#issuecomment-1708513197

     .. code-block:: shell

        export DIST_EXTRA_CONFIG=/tmp/setuptools-build.cfg
        echo -e '[build_ext]\nparallel = 8\n[bdist_wheel]\npy_limited_api = cp311' > $DIST_EXTRA_CONFIG
        python -m build


"""

import logging
import os
import sys
from pathlib import Path

from ..parser_in import TomlParser
from .config_settings import ConfigSettings
from .hooks.manager import (  # noqa: F401
    after,
    before,
    get_plugin_manager,
)
from .wrap_get_version import (
    SEM_VERSION_FALLBACK_SANE,
    write_to_file,
)

log = logging.getLogger("drain_swamp.monkey.wrap_infer_version")

__all__ = ("run_build_plugins",)


def inspect_pm(pm) -> None:  # pragma: no cover
    """Determine which plugin implementations match the spec

    :param pm: Plugin manager instance
    :type pm: pluggy.PluginManager

    .. seealso::

       https://pluggy.readthedocs.io/en/latest/#inspection
    """
    mod_path = "drain_swamp.monkey.wrap_infer_version:inspect_pm"

    wo_specs = []
    for name in pm.hook.__dict__.items():
        if name[0] != "_":
            # : HookCaller
            hook = getattr(pm.hook, name[0])
            msg_info = (
                f"{mod_path} hook.get_hookimpls {name[0]} -> {hook.get_hookimpls()}"
            )
            log.info(msg_info)
            if not hook.has_spec():
                wo_specs.append(name)
    msg_info = f"{mod_path} wo spec hooks {wo_specs}"
    log.info(msg_info)


def run_build_plugins(d_config_settings):
    """Run build plugins. The plugins are responsible for user input
    validation and choosing which user input is of interest

    :param d_config_settings:

       config settings dict. What normally would be supplied as
       :code:`python -m build` config setting cli options

    :type d_config_settings: collections.abc.Mapping[str, typing.Any]
    """
    mod_path = "drain_swamp.monkey.wrap_infer_version:run_build_plugins"

    # late load module
    import drain_swamp.monkey.plugins as package_plugins

    # Can register more pluggy specs with _register_hooks
    # If positional arg not a types.ModuleType raises TypeError
    pm = get_plugin_manager(package_plugins)

    # Very useful check to see if specs fcn names match hook fcn names
    # raises pluggy.PluginValidationError
    # pm.check_pending()
    pass

    # https://pluggy.readthedocs.io/en/latest/api_reference.html#pluggy.PluginManager.add_hookcall_monitoring
    # undo = pm.add_hookcall_monitoring(before, after)
    pass

    result_lst_before = pm.hook.ds_before_version_infer(
        config_settings=d_config_settings
    )
    result_lst_on = pm.hook.ds_on_version_infer(config_settings=d_config_settings)
    result_lst_after = pm.hook.ds_after_version_infer(config_settings=d_config_settings)

    # print hook results. Group by hook spec. Plugins executed in order
    if len(result_lst_before) != 0:  # pragma: no cover
        str_msgs = "\n".join(result_lst_before)
        msg = f"{mod_path} plugin version_infer (before): {str_msgs}"
        log.warning(msg)

    if len(result_lst_on) != 0:  # pragma: no cover
        str_msgs = "\n".join(result_lst_on)
        msg = f"{mod_path} plugin version_infer (on): {str_msgs}"
        log.warning(msg)

    if len(result_lst_after) != 0:  # pragma: no cover
        str_msgs = "\n".join(result_lst_after)
        msg = f"{mod_path} plugin version_infer(after): {str_msgs}"
        log.warning(msg)

    # undo()
    pass


def _rage_quit(is_maybe_none, msg):
    """Check if None otherwise raise SystemExit exception
    :param is_maybe_none:
    :type is_maybe_none: typing.Any | None
    :param msg: Log message if condition is False
    :type msg: str
    :raises:

       - :py:exc:`SystemExit` -- No point in continuing

    """
    is_definitely_none = is_maybe_none is None
    if is_definitely_none:
        log.warning(msg)
        sys.exit(1)
    else:
        pass


def infer_version(dist):
    """Sets the dist.metadata.version

    Would also like to:

    - get cmdline options -C--kind and -C--set-lock
    - refresh_links
    - write to  version_file. kind: tag, current, now, and version str

    In ``pyproject.toml``

    .. code-block:: text

       [project.entry-points."setuptools.finalize_distribution_options"]
       drain_swamp = "drain_swamp.monkey.wrap_infer_version:infer_version"

    :param dist: API interface for interacting with setuptools
    :type dist: setuptools.Distribution

    .. todo:: remove restriction

       _req_links/backend.py should be refactored use the plugins

    .. seealso::

       ``_req_links/backend.py``

    """
    # In the future, remove this restriction
    dist_name = dist.metadata.name
    is_drain_swamp = dist_name is not None and dist_name == "drain_swamp"
    if is_drain_swamp:  # pragma: no cover
        return
    else:  # pragma: no cover
        pass

    # Until setuptools implements passing config-settings into subprocess
    cs = ConfigSettings()
    d_config_settings = cs.read()
    del cs

    mod_path = "drain_swamp.monkey.wrap_infer_version::infer_version"

    """Normally called from within a virtual environment (venv).
    Seems to provide environment variable: PWD, OLDPWD

    If not called from within a venv, fallback is cwd
    """
    # print(f"os.environ: {os.environ!r}")
    project_abspath = os.environ.get("PWD", Path.cwd())

    msg = (
        f"{mod_path} virtual environment did not provide PWD "
        "environment variable. Where is package base or pyproject.toml?"
    )
    _rage_quit(project_abspath, msg)

    path_cwd = Path(project_abspath)
    path_f = path_cwd.joinpath("pyproject.toml")

    tp = TomlParser(path_f)

    msg = f"{mod_path} pyproject.toml either not found or parsing had issues"
    _rage_quit(tp.d_pyproject_toml, msg)

    config_path = str(tp.path_file)

    write_to_file(
        config_path,
        SEM_VERSION_FALLBACK_SANE,
        is_only_not_exists=True,
    )
    run_build_plugins(d_config_settings)



from collections.abc import Callable
from typing import (
    Any,
    TypeVar,
    overload,
)

_F = TypeVar("_F", bound=Callable[..., object])

# From types-pluggy
class HookspecMarker:
    project_name: str
    def __init__(self, project_name: str) -> None: ...
    @overload
    def __call__(
        self,
        function: _F,
        firstresult: bool = False,
        historic: bool = False,
        warn_on_impl: Warning | None = None,
    ) -> _F: ...
    @overload
    def __call__(
        self,
        function: None = None,
        firstresult: bool = False,
        historic: bool = False,
        warn_on_impl: Warning | None = None,
    ) -> Callable[[_F], _F]: ...

hook_spec: HookspecMarker  # noqa: E305

@hook_spec
def ds_before_version_infer(config_settings: dict[str, Any]) -> str | None: ...
@hook_spec
def ds_on_version_infer(config_settings: dict[str, Any]) -> str | None: ...
@hook_spec
def ds_after_version_infer(config_settings: dict[str, Any]) -> str | None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

**Update changelog, NOTICE.txt and ``docs?/conf.py``**

.. code-block:: shell

   igor seed
   igor edits

1. Writes a placeholder block to CHANGES.rst
2. Edits:

   - doc?/conf.py
   - NOTICE.txt
   - CHANGES.rst

**Build package**

Running these commands will also update src/[package name]/_version.py

.. code-block:: shell

   python src/drain_swamp/cli_igor.py build --kind="0.0.1"

or

.. code-block:: shell

   python -m drain_swamp.cli_igor build --kind="0.0.1"

First first form can be run from source, without the package being installed

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

.. py:data:: help_path
   :type: str

   cli option ``--path`` doc string

.. py:data:: help_kind
   :type: str

   cli option ``--kind`` doc string

.. py:data:: help_snippet_co
   :type: str

   cli option --snip doc string

.. py:data:: EPILOG_SEED
   :type: str

   Exit codes explanation for command, ``seed``

.. py:data:: EPILOG_EDITS
   :type: str

   Exit codes explanation for command, ``edits``

.. py:data:: EPILOG_BUILD
   :type: str

   Exit codes explanation for command, ``build``

.. py:data:: EPILOG_PRETAG
   :type: str

   Exit codes explanation for command, ``pretag``

"""

import contextlib
import io
import logging
import sys
from pathlib import Path

import click

# pep366 ...
# https://stackoverflow.com/a/34155199
if __name__ == "__main__" and __spec__ is None:  # pragma: no cover
    # Package not installed
    # python src/drain_swamp/cli_igor.py build --kind="0.0.1"
    import importlib.util

    path_d = Path(__file__).parent
    rev_mods = []
    while path_d.joinpath("__init__.py").exists():
        name = path_d.name
        path_prev = path_d
        path_d = path_d.parent
        rev_mods.append(name)
    # One level above top package --> sys.path.insert
    sys.path.insert(1, str(path_d))
    # parent (aka package) dotted path
    dotted_path = ".".join(reversed(rev_mods))

    # print(f"str(path_d): {str(path_d)}", file=sys.stderr)
    # print(f"dotted_path: {dotted_path}", file=sys.stderr)
    # print(f"path_prev: {path_prev}", file=sys.stderr)
    pass

    # import top package level
    path_top = path_prev.joinpath("__init__.py")
    spec_top = importlib.util.spec_from_file_location(name, path_top)
    mod_top = importlib.util.module_from_spec(spec_top)
    sys.modules[dotted_path] = mod_top
    spec_top.loader.exec_module(mod_top)

    # __spec__ is None. Set __spec__ rather than :code:`__package__ = dotted_path`
    dotted_path_this = f"{dotted_path}.{Path(__file__).stem}"
    spec_this = importlib.util.spec_from_file_location(dotted_path_this, Path(__file__))
    __spec__ = spec_this
elif (
    __name__ == "__main__" and isinstance(__package__, str) and len(__package__) == 0
):  # pragma: no cover
    # When package is not installed
    # python -m drain_swamp.cli_igor build --kind="0.0.1"
    tmp_pkg = "drain_swamp"
    path_pkg_base_dir = Path(__file__).parent.parent
    sys.path.insert(1, str(path_pkg_base_dir))
    mod = __import__(tmp_pkg)
    sys.modules[tmp_pkg] = mod
    __package__ = tmp_pkg
else:
    # normal import
    # __package__ = "drain_swamp"
    pass
# pep366 ...done

from .constants import g_app_name
from .igor_utils import (
    build_package,
    edit_for_release,
    get_current_version,
    get_tag_version,
    pretag,
    print_cheats,
    seed_changelog,
    write_version_file,
)
from .snip import (
    ReplaceResult,
    Snip,
)
from .snippet_sphinx_conf import SnipSphinxConf

_logger = logging.getLogger(f"{g_app_name}.cli_igor")
# taken from pyproject.toml
entrypoint_name = "drain-swamp"  # noqa: F401

help_path = "package root folder"
help_snippet_co = (
    "Snippet code, within a file, unique id of an editable region, aka snippet. "
    "Only necessary if allows for multiple snippets"
)
help_kind = (
    "version string kind: now (alias of current), current, tag, or "
    "explicit semantic version"
)
help_package_name = (
    "Setuptools-scm environment variable contains the package name. "
    """When --kind is "tag", and not provided, queries git. Better to avoid"""
)

EPILOG_SEED = """
EXIT CODES

0 -- always

"""

EPILOG_EDITS = """
EXIT CODES

0 -- Evidently sufficient effort put into unittesting. Job well done, beer on me!

1 -- no doc?/conf.py folder or file

2 -- semantic version str is malformed

"""

EPILOG_LIST = """
EXIT CODES

0 -- Printed snippet codes list and list of snippet code with respective block

3 -- Expected a doc/ or docs/ folder

4 -- Expected to find file, doc/conf.py or docs/conf.py

5 -- Snippet validation fail. Either nested or non-matching start/end tokens

6 -- There are no snippets

"""

EPILOG_BUILD = """
EXIT CODES

0 -- Package built and placed in dist/

1 -- Package build failed. Should print entire build output

2 -- Expecting current working folder to be project base folder. Not a folder

4 -- Could not get package name from pyproject.toml, needed to set the correct version

5 -- Not valid a semantic version str. Only occurs when a version str is provided

"""

EPILOG_SCM_PAIR = """

0 -- Package built and placed in dist/

2 -- Expecting current working folder to be project base folder. Not a folder

4 -- Could not get package name from pyproject.toml, needed to set the correct version

5 -- Not valid a semantic version str. Only occurs when a version str is provided

"""

EPILOG_PRETAG = """
EXIT CODES

0 -- Sanitized semantic version str

1 -- Error with semantic version str. Cannot be fixed

"""

EPILOG_CURRENT_VERSION = """

0 -- Current semantic version str

1 -- Most likely setuptools-scm package is not installed otherwise command failed

"""

EPILOG_TAG_VERSION = """

0 -- Tag semantic version str from version file or fallback current version

1 -- Most likely setuptools-scm package is not installed otherwise command failed

4 -- --kind nonsense. Explicit version str invalid

6 -- Could not get package name from git

"""

EPILOG_CHEATS = """

0 -- printed the cheats

2 -- Unknown parameter or parameter has unsupported type

3 -- Expecting a folder, got something else

4 -- --kind nonsense. Explicit version str invalid

6 -- Could not get package name from git

"""


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
)
def main():
    """Command-line for drain-swamp. Prints usage"""


@main.command(
    "seed",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_SEED,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
def seed(path):
    """Updates changelog, ``CHANGES.rst``, creating placeholder

    If file [path]/CHANGES.rst does not exist, a warning is logged,
    need to capture that log entry

    \f
    :param path: path to the current working directory containing pyproject.toml
    :type path: pathlib.Path
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    with contextlib.redirect_stderr(io.StringIO()) as f_stream:
        seed_changelog(path)
    str_err = f_stream.getvalue()
    click.echo(str_err)


@main.command(
    "edits",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_EDITS,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-k",
    "--kind",
    "kind",
    default="tag",
    type=click.STRING,
    help=help_kind,
)
@click.option(
    "-s",
    "--snip",
    "snippet_co",
    default=None,
    type=click.STRING,
    help=help_snippet_co,
)
def edit(path, kind, snippet_co=None):  # pragma: no cover
    """Edits: doc?/conf.py, NOTICE.txt, and CHANGES.rst

    \f
    :param path: path to the current working directory containing pyproject.toml
    :type path: pathlib.Path
    :param kind:

       semantic version str or "current" or "now" or "tag". Side effect
       changes ``src/[app name]/_version.py``

    :type kind: str
    :param snippet_co: Sphinx doc?/conf.py snippet code. Supply none is no snippet has no code
    :type snippet_co: str | None
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    with contextlib.redirect_stderr(io.StringIO()) as f_stream:
        opt_int = edit_for_release(path, kind, snippet_co=snippet_co)
    str_err = f_stream.getvalue()
    click.echo(str_err, err=True)

    if opt_int is None:
        sys.exit(0)
    else:
        sys.exit(opt_int)


@main.command(
    "list",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_LIST,
)
@click.option(
    "-p",
    "--path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
def snippets_list(path):
    """In Sphinx doc?/conf.py, list snippets

    path is the package base folder. Searches for ``doc?/conf.py``

    .. code-block shell:

       drain-swamp list
       cd tests
       drain-swamp list --path=..

    \f
    :param path:

       Default current working directory. package root folder. Must
       contain ``pyproject.toml``

    :type path: pathlib.Path
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    # Get absolute path to doc?/conf.py
    try:
        sc = SnipSphinxConf(path=path)
    except NotADirectoryError:
        msg_exc = "Expected a doc/ or docs/ folder"
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(3)
    except FileNotFoundError:
        msg_exc = "Expected to find file, doc/conf.py or docs/conf.py"
        click.secho(msg_exc, fg="red", err=True)
        sys.exit(4)

    abspath_conf_py = sc.path_abs

    # print:
    # - snippet codes
    # - snippets
    snip = Snip(abspath_conf_py)
    with contextlib.redirect_stderr(io.StringIO()) as f:
        snippets = snip.print()
    msg = f.getvalue()
    if isinstance(snippets, ReplaceResult):
        click.secho(msg, fg="red", err=True)
        if snippets == ReplaceResult.VALIDATE_FAIL:
            sys.exit(5)
        elif snippets == ReplaceResult.NO_MATCH:
            sys.exit(6)
        else:  # pragma: no cover
            pass
    else:
        click.secho(msg, fg="green", err=True)


@main.command(
    "build",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_BUILD,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-k",
    "--kind",
    "kind",
    default="tag",
    type=click.STRING,
    help=help_kind,
)
def semantic_version_aware_build(path, kind):
    """Build package

    \f
    :param path: current working directory
    :type path: pathlib.Path
    :param kind:

       semantic version str or "current" or "now" or "tag". Side effect
       changes ``src/[app name]/_version.py``

    :type kind: str
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    try:
        with contextlib.redirect_stderr(io.StringIO()) as f_stream:
            # click --> exit code 2. Preventing NotADirectoryError
            bol_out = build_package(path, kind)
    except AssertionError as e:
        # AssertionError
        msg_exc = str(e)
        click.echo(msg_exc, err=True)
        sys.exit(4)
    except ValueError as e:
        msg_exc = str(e)
        click.echo(msg_exc, err=True)
        sys.exit(5)
    if not bol_out:
        # need entire output to know how to deal with the failure
        str_err = f_stream.getvalue()
        click.echo(str_err, err=True)
        sys.exit(1)
    else:
        sys.exit(0)


@main.command(
    "write_version",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_SCM_PAIR,
    deprecated=True,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-k",
    "--kind",
    "kind",
    default="tag",
    type=click.STRING,
    help=help_kind,
)
def setuptools_scm_key_value_pair(path, kind):
    """Given kind, write version str to version_file

    \f
    :param path: current working directory
    :type path: pathlib.Path
    :param kind:

       semantic version str or "current" or "now" or "tag". Side effect
       changes ``src/[app name]/_version.py``

    :type kind: str

    .. deprecated:: 0.5.1

       Version file written by plugin, ds_scm_version, during sdist
       build. Might remain relevent only to initially create the version file

    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    try:
        with contextlib.redirect_stderr(io.StringIO()):
            # click --> exit code 2. Preventing NotADirectoryError
            write_version_file(path, kind)
    except AssertionError as e:
        # AssertionError
        msg_exc = str(e)
        click.echo(msg_exc, err=True)
        sys.exit(4)
    except ValueError as e:
        msg_exc = str(e)
        click.echo(msg_exc, err=True)
        sys.exit(5)


@main.command(
    "pretag",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_PRETAG,
)
@click.argument("ver", type=click.STRING)
def validate_tag(ver):
    """Print the sanitized semantic version str

    Normal cli usage

    drain-swamp pretag "0.0.1"

    Usage when package is not installed

    python src/drain_swamp/cli_igor.py pretag "0.0.1"

    Usage if package is installed

    python -m drain_swamp.cli_igor pretag "0.0.1"

    \f
    :param ver: Possibility malformed semantic ver str
    :type ver: str
    """
    is_success, sanitized = pretag(ver)
    click.echo(sanitized)
    if is_success:
        sys.exit(0)
    else:
        sys.exit(1)


@main.command(
    "current",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_CURRENT_VERSION,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
def current_version(path):
    """Get scm version str

    python src/drain_swamp/cli_igor.py current

    or

    drain-swamp current

    \f
    :param path: current working directory
    :type path: pathlib.Path
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    opt_str = get_current_version(path)
    if opt_str is None:
        sys.exit(1)
    else:
        click.echo(opt_str)
        sys.exit(0)


@main.command(
    "tag",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_TAG_VERSION,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    help=help_path,
)
def tag_version(path):
    """Get semantic version str from version_file. Fall back to current version

    python src/drain_swamp/cli_igor.py tag

    or

    drain-swamp tag

    \f
    :param path: current working directory
    :type path: pathlib.Path
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    # NotADirectoryError prevents by click
    try:
        ver_sem = get_tag_version(path)
    except ValueError as e:
        """tag version skipped if version file semantic version str is
        invalid. Then if current version is invalid --> ValueError"""
        msg = str(e)
        exit_code = 4
    except AssertionError as e:
        # pyproject.toml file missing or malformed
        msg = str(e)
        exit_code = 6
    else:
        msg = None
        exit_code = 0

    if msg is not None:
        click.secho(msg, fg="green", err=True)
        sys.exit(exit_code)
    else:  # pragma: no cover
        click.echo(ver_sem)


@main.command(
    "cheats",
    context_settings={"ignore_unknown_options": True},
    epilog=EPILOG_CHEATS,
)
@click.option(
    "-p",
    "--path",
    "path",
    default=Path.cwd(),
    type=click.Path(exists=False, file_okay=True, dir_okay=True, resolve_path=True),
    help=help_path,
)
@click.option(
    "-k",
    "--kind",
    "kind",
    default="tag",
    type=click.STRING,
    help=help_kind,
)
def do_cheats(path, kind):
    """Get useful notes to aid in kitting and publishing

    python src/drain_swamp/cli_igor.py cheats

    or

    drain-swamp cheats

    \f
    :param path: current working directory
    :type path: pathlib.Path
    :param kind:

       semantic version str or "current" or "now" or "tag". Side effect
       changes ``src/[app name]/_version.py``

    :type kind: str
    """
    # resolve causing conversion into a str. Should be Path
    if isinstance(path, str):  # pragma: no cover
        path = Path(path)
    else:  # pragma: no cover
        pass

    try:
        print_cheats(path, kind)
    except NotADirectoryError as e:
        # Expecting a folder, got something else
        msg = str(e)
        exit_code = 3
    except ValueError as e:
        # --kind nonsense. Explicit version str invalid
        msg = str(e)
        exit_code = 4
    except AssertionError as e:
        # Could not get package name from git
        msg = str(e)
        exit_code = 6
    else:
        msg = None
        exit_code = 0

    if msg is not None:
        click.secho(msg, fg="green", err=True)
    else:  # pragma: no cover
        pass

    sys.exit(exit_code)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

git --> setuptools-scm --> kitting (howto.txt / igor.py / Makefile) --> semantic versioning

.. seealso::

   packaging.version.Version
   `[docs] <https://packaging.pypa.io/en/stable/version.html>`_

Release phases
---------------

Without SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP environmental variable
locals are included in version

e.g. "0.1.1.dev0+g4b33a80.d20240129" local is "g4b33a80.d20240129"

When releasing this is not what is wanted, so use
SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP with the version

- Current version

  .. code-block:: shell

     PYTHONWARNINGS="ignore" python setup.py --version


- Release by tag aka final

  .. code-block:: shell

     PYTHONWARNINGS="ignore" SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="$(git describe --tag)" python setup.py --version
     SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="$(git describe --tag)" python -m build


- alpha: a, beta: b, or candidate: rc

  .. code-block:: shell

     PYTHONWARNINGS="ignore" SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="0.1.1a1" python setup.py --version
     SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="0.1.1a1" python -m build


- dev

  .. code-block:: shell

     PYTHONWARNINGS="ignore" SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="0.1.1a1.dev1" python setup.py --version
     SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="0.1.1a1.dev1" python -m build


Move the tag past post commits

- post

  .. code-block:: shell

     PYTHONWARNINGS="ignore" SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="0.1.1.post1" python setup.py --version
     SETUPTOOLS_SCM_PRETEND_VERSION_FOR_DRAIN_SWAMP="0.1.1.post1" python -m build


.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("SemVersion", "sanitize_tag", "get_version")

   Module exports

.. py:data:: _map_release
   :type: types.MappingProxyType[str, str]
   :value: types.MappingProxyType({"alpha": "a", "beta": "b", "candidate": "rc"})

   Mapping of release levels. So can gracefully go back and forth

   Read only mapping. key is long name. value is abbreviation. Long
   names will be converted into the abbreviations

"""

import importlib.util
import logging
import sys
import types
import warnings
from collections.abc import Sequence
from functools import partial
from pathlib import (
    Path,
    PurePath,
)

try:
    from packaging.version import InvalidVersion
    from packaging.version import Version as Version
except ImportError:  # pragma: no cover
    from setuptools.extern.packaging.version import InvalidVersion  # type: ignore
    from setuptools.extern.packaging.version import Version as Version  # type: ignore

from ._run_cmd import run_cmd
from .parser_in import TomlParser
from .version_file._overrides import (
    PRETEND_KEY_NAMED,
    normalize_dist_name,
)

__package__ = "drain_swamp"
__all__ = (
    "SemVersion",
    "sanitize_tag",
    "get_version",
)

_map_release = types.MappingProxyType({"alpha": "a", "beta": "b", "candidate": "rc"})
_logger = logging.getLogger("drain_swamp.version_semantic")


def _is_package(package_name):
    """Check if package installed

    :param package_name: package name to check if installed
    :type package_name: str
    :returns: True if package installed otherwise False
    :rtype: str
    :meta private:
    """
    from importlib.metadata import (
        PackageNotFoundError,
        version,
    )

    try:
        version("setuptools-scm")
    except PackageNotFoundError:
        bol_ret = False
    else:
        bol_ret = True

    return bol_ret


def _is_setuptools_scm():
    """Without importing package, check setuptools-scm package is installed

    :returns: True if setuptools-scm package is installed otherwise False
    :rtype: bool
    :meta private:
    """
    ret = _is_package("setuptools-scm")

    return ret


def _path_or_cwd(val):
    """Frequently used and annoying to test multiple times

    Should be a Path. If anything else return cwd

    :param path: Should be a Path
    :type path: typing.Any | None
    :returns: Path or cwd Path
    :rtype: pathlib.Path
    :meta private:
    """
    if val is None or not issubclass(type(val), PurePath):
        path_cwd = Path.cwd()
    else:
        path_cwd = val

    return path_cwd


def get_package_name(path):
    """Get package name, unmodified, from pyproject.toml

    :param path: absolute path to either package base folder or ``pyproject.toml``
    :type path: pathlib.Path
    :returns:

       package name. ``[project].name`` is a required field, so can
       assume exists. None if issue with ``pyproject.toml``

    :rtype: str | None
    """
    path_ = _path_or_cwd(path)
    tp = TomlParser(path_)
    d_pyproject_toml = tp.d_pyproject_toml
    if d_pyproject_toml is not None:
        ret = d_pyproject_toml.get("project", {}).get("name", None)
    else:
        ret = None

    return ret


def _scm_key(dist_name):
    """When want to set a specific version, setuptools-scm offers an
    environment variable which overrides normal behavior

    This is needed when wanting to create a:
    - tagged version
    - post-release version
    - pre-release version

    :param prog_name:

       package name. Will upper case and replace hyphens with underscores

    :type prog_name: str
    :returns: environment variable to affect setuptools-scm behavior
    :rtype: str
    :meta private:
    """
    # source setuptools_scm._override.read_named_env
    env_var_dist_name = normalize_dist_name(dist_name)
    d_named = {"name": env_var_dist_name}
    scm_override_key = PRETEND_KEY_NAMED.format(**d_named)

    return scm_override_key


def _current_tag(path=None):
    """Tagged version as known by git. If there is no tagged version, make one!

    Until a tagged version exists, the version string will not be a
    valid semantic version.

    The git version is authoritative, **not** some hardcoded version str
    in a module within a Python package

    Runs :command:`git describe --tag`

    :param path:

       Default None. If None assumes path is current working directory,
       otherwise provide the path to the package base folder

    :type path: pathlib.Path | None
    :returns:

       None indicates there are no tagged versions otherwise the latest tagged version

    :rtype: str | None
    :meta private:
    """
    path_cwd = _path_or_cwd(path)

    cmd = ["/bin/git", "describe", "--tag"]
    t_ret = run_cmd(cmd, cwd=path_cwd)
    out, err, exit_code, exc = t_ret
    if exc is not None:
        # git says, no tagged version
        str_out = None
    else:
        # No tags yet
        # 128 -- fatal: No names found, cannot describe anything.
        is_fail = exit_code != 0
        if not is_fail:
            str_out = out
        else:
            str_out = None

    return str_out


def _strip_epoch(ver):
    """Strip epoch

    :param ver: May contain epoch, ``v``, and local
    :type ver: str
    :returns: epoch and remaining str including ``v`` and local
    :rtype: tuple[str | None, str]
    :meta private:
    """
    try:
        idx = ver.index("!")
    except ValueError:
        # Contains no epoch
        epoch = None
        remaining = ver
    else:
        epoch = ver[: idx + 1]
        remaining = ver[idx + 1 :]

    return epoch, remaining


def _strip_local(ver):
    """Strip local from end of version string

    From ``0!v1.0.1.a1dev1+g4b33a80.d20240129``

    local: ``g4b33a80.d20240129``

    remaining: ``0!v1.0.1.a1dev1``

    :param ver: full version str from git. May include epoch and local
    :type ver: str
    :returns: local and remaining
    :rtype: tuple[str | None, str]
    :meta private:
    """
    try:
        idx = ver.index("+")
    except ValueError:
        # Contains no local
        local = None
        remaining = ver
    else:
        local = ver[idx + 1 :]
        remaining = ver[:idx]

    return local, remaining


def _remove_v(ver):
    """Remove prepended v. e.g. From ``0!v1.0.1.a1dev1+g4b33a80.d20240129``

    Will not work on an initial untagged version, ``0.1.dev0.d20240213``

    :param ver: Non-initial untagged version
    :type ver: str
    :returns: original str without the v. Includes epoch and local
    :rtype: str
    :meta private:
    """
    epoch, remaining = _strip_epoch(ver)
    local, remaining = _strip_local(remaining)

    # If prepended "v", remove it. epoch e.g. ``1!v1.0.1`` would conceal the "v"
    if remaining.startswith("v"):
        remaining = remaining[1:]

    ret = epoch if epoch is not None else ""
    ret += remaining

    if local is not None:
        ret += f"+{local}"
    else:  # pragma: no cover
        pass

    return ret


def sanitize_tag(ver):
    """Avoid reinventing the wheel, leverage Version

    ``final`` is not valid

    :param ver: raw semantic version
    :type ver: str
    :returns: Sanitized semantic version str
    :rtype: tuple[str, str | None]
    :raises:

       - :py:exc:`ValueError` -- Invalid token within Version str

    """
    # Careful! Will choke on initial untagged version, e.g. ``0.1.dev0.d20240213``
    str_remaining_whole = _remove_v(ver)

    # Strip epoch, if exists
    _, str_remaining_stripped = _strip_epoch(str_remaining_whole)

    """Strip local, if exists

    Will fail to detect an initial untagged version e.g. '0.1.dev0.d20240213'"""
    local, str_remaining_stripped = _strip_local(str_remaining_stripped)

    # Problematic
    # '0.1.dev0.d20240213'. Untagged version. Try remove from last period
    # 0.1.1.candidate1dev1+g4b33a80.d20240129
    try:
        v = Version(str_remaining_whole)
    except InvalidVersion:
        is_problem = True
    else:
        is_problem = False

    if is_problem:
        lst = str_remaining_whole.split(".")

        ver_try = ".".join(lst[:-1])
        try:
            v = Version(ver_try)
        except InvalidVersion:
            is_still_issue = True
        else:  # pragma: no cover Do nothing
            is_still_issue = False
    else:  # pragma: no cover Do nothing
        is_still_issue = False

    if is_still_issue:
        try:
            v = Version(str_remaining_whole)
        except InvalidVersion as e:
            msg = f"Version contains invalid token. {e}"
            raise ValueError(msg) from e
    else:  # pragma: no cover Do nothing
        pass

    ret = str(v)

    # Strip epoch and local, if exists
    _, ret = _strip_epoch(ret)
    _, ret = _strip_local(ret)

    return ret, local


def get_version(ver, is_use_final=False):
    """Semantic version string broken into parts

    :param ver: A semantic version string
    :type ver: str
    :param is_use_final:

       Default False. ``final`` is not normally valid within a semantic
       version string. The use of final may be used to indicate intention of creating
       a tagged release version. If all the stars are aligned and its
       G'ds Will. If not, ``post release`` version(s) would occur and
       ``final`` would be incorrect.

       Don't create invalid semantic version strings cuz its convenient.
       Don't use this feature!

    :type is_use_final: bool
    :returns:

       Semantic version broken into parts: major, minor, micro,
       release level, serial. And _dev

    :rtype: tuple[tuple[int, int, int, str, int], int | None]
    :raises:

       - :py:exc:`ValueError` -- Invalid version string. git
         requires one commit. setupttools-scm requires one tagged commit

    """
    if is_use_final is None or not isinstance(is_use_final, bool):
        is_use_final = False
    else:  # pragma: no cover
        pass

    # epoch and locals and prepend v ... remove
    _v = _remove_v(ver)

    try:
        _v = Version(_v)
    except InvalidVersion as e:
        msg = f"Version contains invalid token. {e}"
        raise ValueError(msg) from e

    _dev = _v.dev if _v.is_devrelease else None

    if not _v.is_prerelease and not _v.is_postrelease:
        # ``final`` means intend to bump version. Not actually valid
        releaselevel = "" if not is_use_final else "final"
        serial = 0
        _dev = None
    elif _v.is_prerelease and not _v.is_postrelease:
        # Requires long
        if _v.is_devrelease and _v.pre is None:
            # dev
            serial = 0
            releaselevel = ""  # alpha??
        else:
            # alpha beta, candidate, a, b, or rc
            t_pre = _v.pre
            short = t_pre[0]
            serial = t_pre[1]
            for long_, short_ in _map_release.items():
                if short_ == short:
                    releaselevel = long_
                else:  # pragma: no cover continue
                    pass
    elif not _v.is_prerelease and _v.is_postrelease:
        releaselevel = "post"
        serial = _v.post
    elif _v.is_prerelease and _v.is_postrelease:  # pragma: no cover
        # impossible
        pass
    else:  # pragma: no cover
        pass

    return (_v.major, _v.minor, _v.micro, releaselevel, serial), _dev


def _current_version(path=None):
    """:py:mod:`setuptools_scm` get the current
    version. Which more often than not is a development version

    :param path:

       Default None. If None assumes path is current working directory,
       otherwise provide the path to the package base folder

    :type path: pathlib.Path | None
    :returns:

       Current version as known by setuptools-scm. Avoid returning an empty string

    :rtype: str | None
    :meta private:

    .. note::

       To update ``src/[app name]/_version.py``, use command,
       :code:`python setup.py --version`

    """
    path_cwd = _path_or_cwd(path)

    # Check setuptools-scm package is installed
    if not _is_setuptools_scm():
        return None
    else:  # pragma: no cover
        pass

    # Call in subprocess to avoid circular import error
    # Replaces:
    # cmd = [sys.executable, "setup.py", "--version"]
    # cmd = [sys.executable, "-m", "setuptools_scm"]
    if _is_package("drain_swamp"):
        # thru entrypoint
        cmd = ("scm-version", "get")
    else:  # pragma: no cover
        # thru pep366. unittest adds nothing
        cmd = (sys.executable, "src/drain_swamp/cli_scm_version.py", "get")

    t_ret = run_cmd(cmd, cwd=path_cwd)
    out, err, exit_code, exc = t_ret
    if exc is not None:
        # scm-version get says no tagged version
        str_out = None
    else:
        is_fail = exit_code != 0
        if not is_fail:
            str_out = out
        else:
            str_out = None

    return str_out


def _tag_version(
    next_version="",
    path=None,
):
    """Previously meant latest git tagged version, but there is no
    situation where this is desirable.

    If ``tag`` version, take from: version_file, or fallback
    If version provided, update version file

    :param next_version: Default empty string. If not provided, tagged version
    :type next_version: str | None
    :param path:

       Default None. If None assumes path is current working directory,
       otherwise provide the path to the package base folder

    :type path: pathlib.Path | None
    :returns: tagged version or current version str
    :rtype: str | None
    :meta private:
    :raises:

       - :py:exc:`AssertionError` -- ``pyproject.toml`` missing or unparsable

    """
    # empty str means take current tag version
    ret = _arbritary_version(next_version, path=path)

    return ret


def _arbritary_version(
    next_version,
    path=None,
):
    """From the version file get semantic version str. This depends heavily on
    ``pyproject.toml`` being well-formed

    :param next_version: Default empty string. If not provided, tagged version
    :type next_version: str
    :param path:

       Default None. If None assumes path is current working directory,
       otherwise provide the path to the package base folder

    :type path: pathlib.Path | None
    :returns: tagged version or current version str. None rather than empty string
    :rtype: str | None
    :meta private:

    .. warning::

       This command writes to src/[package name]/_version.py
       Use :py:func:`unittest.mock.patch` to avoid the actual call

    """
    msg_issue = "Could not get tag version from version_file"
    path_cwd = _path_or_cwd(path)

    tp = TomlParser(path_cwd)
    d_pyproject_toml = tp.d_pyproject_toml
    if d_pyproject_toml is None:
        msg_warn = f"Issue with pyproject.toml, {msg_issue}"
        _logger.warning(msg_warn)
        raise AssertionError(msg_warn)
    else:  # pragma: no cover
        path_f = tp.path_file
        path_package_base = path_f.parent

    """From here, for all issues, issue a non-blocking warning and return None.
    This strategy is so can fallback to the current version"""
    lst_dynamic = d_pyproject_toml.get("project", {}).get("dynamic", [])
    is_version_file = isinstance(lst_dynamic, Sequence) and "version" in lst_dynamic
    static_ver = d_pyproject_toml.get("project", {}).get("version", None)
    is_version_static = (
        static_ver is not None
        and isinstance(static_ver, str)
        and len(static_ver.strip()) != 0
    )
    if is_version_static:
        # static version
        ver_tag = d_pyproject_toml.get("project", {}).get("version", None)
    elif is_version_file:
        # dynamic version_file
        d_dynamic = (
            d_pyproject_toml.get("tool", {}).get("setuptools", {}).get("dynamic", {})
        )

        is_version = (
            "version" in d_dynamic.keys()
            and isinstance(d_dynamic["version"], dict)
            and (
                "attr" in d_dynamic["version"].keys()
                or "file" in d_dynamic["version"].keys()
            )
        )

        # Check tool.setuptools.dynamic.version
        if not is_version:  # pragma: no cover
            msg_warn = (
                "Expecting tool.setuptools.dynamic.version to contain "
                f"attr or file key, {msg_issue}"
            )
            warnings.warn(msg_warn)
            return None
        else:  # pragma: no cover
            pass

        d_version = d_dynamic["version"]
        is_attr = "attr" in d_version.keys() and isinstance(d_version["attr"], str)
        is_file = "file" in d_version.keys() and isinstance(d_version["file"], str)
        if is_attr:
            # https://setuptools.pypa.io/en/latest/userguide/pyproject_config.html
            # ast.literal_eval() ??
            # attr is dotted path to module variable containing semantic version str
            # file would be to a text file
            version_file_dottedpath = d_dynamic["version"]["attr"]
            dots = version_file_dottedpath.split(".")
            variable_name = dots[-1]
            dotted_path_to_module = dots[:-1]

            # this is an assumption
            # ############################################################################################
            lst_relpath = ["src"]

            lst_relpath.extend(dotted_path_to_module)

            # Coming from a dotted_path, restore the suffix
            file_stem = Path(lst_relpath[-1]).stem
            lst_relpath[-1] = f"{file_stem!s}.py"

            path_version_file = path_package_base.joinpath(*lst_relpath)
            is_not_file = not (
                path_version_file.exists()
                and path_version_file.is_file()
                and path_version_file.suffixes == [".py"]
            )
            if is_not_file:
                msg_warn = (
                    f"version_file {path_version_file} either does "
                    f"not exist or not a file, {msg_issue}"
                )
                warnings.warn(msg_warn)
                return None
            else:  # pragma: no cover
                name = path_version_file.stem
                spec = importlib.util.spec_from_file_location(
                    name,
                    path_version_file,
                )
                module = importlib.util.module_from_spec(spec)
                loader = importlib.util.LazyLoader(spec.loader)
                loader.exec_module(module)
                mixed_var = getattr(module, variable_name, None)
                if mixed_var is None or not isinstance(mixed_var, str):
                    msg_warn = (
                        "In version_file (python module), "
                        f"{path_version_file} module  variable, "
                        "__version__ expected to be a the semantic "
                        "version str, unsupported type got "
                        f"{type(mixed_var)} {msg_issue}"
                    )
                    warnings.warn(msg_warn)
                    return None
                else:  # pragma: no cover
                    pass
                ver_tag = mixed_var
        elif is_file:
            path_f = path_package_base.joinpath(d_version["file"])
            _logger.info(f"_tag_version dynamic version file: {path_f}")
            is_not_file = not (
                path_f.exists() and path_f.is_file() and path_f.suffixes == [".txt"]
            )
            if is_not_file:
                msg_warn = (
                    "Could not find version_file "
                    f"""{d_version["file"]} {msg_issue}"""
                )
                warnings.warn(msg_warn)
                return None
            else:  # pragma: no cover
                pass

            # limit read size to avoid bufferoverflow attack
            # ###############################################################################################
            ver_tag = path_f.read_text()
        else:  # pragma: no cover
            # Check tool.setuptools.dynamic.version.[file|attr] value valid
            msg_warn = (
                "marked dynamic, tool.setuptools.dynamic.version.[file|attr] "
                f"contains something other than a str. {msg_issue}"
            )
            warnings.warn(msg_warn)
            return None
    else:
        msg_warn = (
            f"In {path_f}, neither static nor dynamic version specified "
            f"pyproject.toml is invalid. {msg_issue}"
        )
        warnings.warn(msg_warn)
        return None

    return ver_tag


def _get_app_name(path=None):
    """app name contains underscores. Package name contains hyphens.
    Want the app name, not the package name

    Source git. Looks at the local repo, not remote

    :param path:

       Default None. If None assumes path is current working directory,
       otherwise provide the path to the package base folder

    :type path: pathlib.Path | None
    :returns: app name. None if fails to get from git
    :rtype: str | None
    :meta private:

    .. seealso::

       `From git get project name <https://stackoverflow.com/a/33180289>`_

    """
    path_cwd = _path_or_cwd(path)

    cmd = (
        "/bin/git",
        "rev-parse",
        "--show-toplevel",
    )
    t_ret = run_cmd(cmd, cwd=path_cwd)
    out, err, exit_code, exc = t_ret
    if exc is not None:
        ret = None
    else:
        if out is None:
            ret = None
        else:
            str_mixed = Path(out).name
            ret = str_mixed.replace("-", "_")

    return ret


class SemVersion:
    """Translates :menuselection:`git version --> Python semantic version`.
    Intended use cases:

    - kitting (building sdist and whl)

    - **not** hardcoding the package str within Python packages

    .. py:attribute:: CURRENT_ALIAS_DEFAULT
       :type: str
       :value: "current"

       Preferred default that is used within code. Indicates to get the
       current git version

    .. py:attribute:: CURRENT_ALIASES
       :type: tuple[str, str]
       :value: ("current", "now")

       Aliases which means current (version)

    .. py:attribute:: KINDS
       :type: tuple[str, str, str]
       :value: ("tag", "current", "now")

       All the acceptable kinds besides a version str

       tag -- get the latest tagged version

       current or now -- get the current version from git. Most
       likely development version

    :ivar path:

       Default None. If None, assumes path is current working directory,
       otherwise provide the absolute path to the package base folder

    :vartype path: pathlib.Path | None
    :ivar is_use_final:

       Default False. ``final`` is not normally valid within a semantic
       version string. The use of final may be used to indicate intention of creating
       a tagged release version. If all the stars are aligned and its
       G'ds Will. If not, ``post release`` version(s) would occur and
       ``final`` would be incorrect.

       Don't create invalid semantic version strings cuz its convenient.
       Don't use this feature!

    :vartype is_use_final: typing.Any | None
    :raises:

       - :py:exc:`NotADirectoryError` -- Not the absolute path to package base folder

    """

    CURRENT_ALIAS_DEFAULT = "current"
    CURRENT_ALIASES = (
        "current",
        "now",
    )
    KINDS = ("tag", "current", "now")

    def __init__(
        self,
        path=None,
        is_use_final=False,
    ):
        super().__init__()

        self.path_cwd = path
        self.is_use_final = is_use_final

    @classmethod
    def sanitize_kind(cls, kind):
        """Allow kind to be a version str, 'current', 'tag'
        :param kind:

           Default None. If None, assumes "tag". Type of desired version str.
           Most reliable to pass in a version str if making a tagged
           version or want the tagged version. "current" will most likely
           get a development version

        :type kind: str | None
        :returns: kind ("current" or "tag") or a version str
        :rtype: str
        """

        def check_is_none(val):
            is_none = val is None
            if is_none:
                ret = "tag"
            else:  # pragma: no cover
                # N/A
                ret = val

            return ret

        def check_accidental_seq_str(val):
            """Accidentily pass in a Sequence[str] rather than an str"""
            is_seq_str_nonempty = (
                val is not None
                and isinstance(val, Sequence)
                and not isinstance(val, str)
                and len(val) != 0
                and isinstance(val[0], str)
            )
            if is_seq_str_nonempty:
                ret = val[0]
            else:
                # N/A
                ret = val

            return ret

        def check_seq_unsupported(val):
            is_seq_str_unsupported = (
                val is not None
                and isinstance(val, Sequence)
                and not isinstance(val, str)
                and len(val) != 0
                and not isinstance(val[0], str)
            )
            if is_seq_str_unsupported:
                ret = "tag"
            else:
                # N/A
                ret = val

            return ret

        def check_str(val):
            if val is not None and isinstance(val, str):
                if val not in cls.KINDS:
                    # Override version str
                    ret = val
                else:
                    if val in cls.CURRENT_ALIASES:
                        # current
                        ret = cls.CURRENT_ALIAS_DEFAULT
                    else:
                        # tag
                        ret = "tag"
            else:  # pragma: no cover
                # N/A
                ret = val

            return ret

        def check_unsupported(val):
            # fallback --> "tag". E.g. kind = 1.2345
            if val != "tag" and val != "current" and not isinstance(val, str):
                ret = "tag"
            else:  # pragma: no cover
                ret = val

            return ret

        kind_0 = check_is_none(kind)
        kind_1 = check_accidental_seq_str(kind_0)
        kind_2 = check_seq_unsupported(kind_1)
        kind_3 = check_str(kind_2)  # if still None
        kind_4 = check_unsupported(kind_3)

        return kind_4

    @property
    def path_cwd(self):
        """Getter for absolute Path to current working directory

        :returns: Absolute Path to current working directory
        :rtype: pathlib.Path
        """
        return self._path_cwd

    @path_cwd.setter
    def path_cwd(self, val):
        """Setter for cwd

        :param val: current working directory absolute Path
        :type val: typing.Any
        :raises:

           - :py:exc:`NotADirectoryError` -- Not the absolute path to package base folder

        """
        path_cwd = _path_or_cwd(val)
        if path_cwd.exists() and path_cwd.is_absolute() and not path_cwd.is_dir():
            msg_exc = "Not the absolute path to package base folder"
            raise NotADirectoryError(msg_exc)
        else:
            self._path_cwd = path_cwd

    @property
    def is_use_final(self):
        """Within a semantic version str "final" is not a valid token.
        However in scripts it might be used to indicate about to create a tagged version

        Use of final is discouraged

        :returns: True if final can be within a semantic version str otherwise False
        :rtype: bool
        """
        return self._is_use_final

    @is_use_final.setter
    def is_use_final(self, val):
        """Setter for is_use_final

        :param val: Should be a bool if not defaults to False
        :type val: typing.Any
        """
        if val is None or not isinstance(val, bool):
            self._is_use_final = False
        else:
            self._is_use_final = val

    @property
    def major(self):
        """major version. If breaking change, in API, should be incremented

        :returns: None if not called parse_ver beforehand. Otherwise will be an int
        :rtype: int | None
        """
        return self._major

    @property
    def minor(self):
        """minor version. Incremented if a new feature or a fix occurred

        :returns: None if not called parse_ver beforehand. Otherwise will be an int
        :rtype: int | None
        """
        return self._minor

    @property
    def micro(self):
        """micro version. For all the other possible categories of
        changes which are not exceptionally notable

        :returns: None if not called parse_ver beforehand. Otherwise will be an int
        :rtype: int | None
        """
        return self._micro

    @property
    def releaselevel(self):
        """Get long form: alpha, beta, candidate, post.
        Abbreviations (a, b, and rc) are converted into long forms:

        Pre-releases: alpha, beta, candidate, a, b, or rc

        Post release: post

        Does not indicate whether a development release or not

        :returns:

           pre-release (beta alpha) or release candidate (rc). None,
           if not call parse_var beforehand

        :rtype: str | None
        """
        return self._releaselevel

    @releaselevel.setter
    def releaselevel(self, val):
        """For release level, convert either short or long form into long form

        :param val: Either short or long release level str
        :type val: str
        """
        ret = val
        for long_, short_ in _map_release.items():
            if short_ == ret:
                ret = long_
            else:  # pragma: no cover continue
                pass
        self._releaselevel = ret

    @property
    def releaselevel_abbr(self):
        """Short form: a, b, rc, post

        :returns: Short form. Valid in semantic version str
        :rtype: str
        """
        long_form = self.releaselevel
        ret = long_form
        for long_, short_ in _map_release.items():
            if long_ == long_form:
                ret = short_

        return ret

    @property
    def serial(self):
        """Whether is a post release or not. Cannot be both a post and a pre release.

        :returns:

           1 indicates post release otherwise 0. None, if not call parse_var beforehand

        :rtype: int | None
        """
        return self._serial

    @property
    def dev(self):
        """Development version number

        :returns:

           dev version number. None is valid. If a dev starts from 0.
           Also None, if not call parse_var beforehand

        :rtype: int | None
        """
        return self._dev

    @property
    def release(self):
        """Components of the release segment of the version

        Does not include epoch or any pre-release / development / post-release
        suffixes

        :returns: tuple of major, minor, micro
        :rtype: tuple[int, int, int]
        """
        return self._release

    @staticmethod
    def as_tuple(version_str):
        """version tuple as written to ``_version.py`` file

        :param version_str: raw version str
        :type version_str: str
        :returns: version tuple
        :rtype: tuple[int | str, ...]
        """
        try:
            ver, local = sanitize_tag(version_str)
        except ValueError:
            return (version_str,)

        sv = SemVersion()
        sv.parse_ver(ver, local=local)

        version_fields: tuple[int | str, ...] = sv.release

        if sv.dev is not None:
            version_fields += (f"dev{sv.dev}",)
        else:  # pragma: no cover
            pass

        if sv._local is not None:
            version_fields += (sv._local,)
        else:  # pragma: no cover
            pass

        return version_fields

    def parse_ver(self, ver, local=None):
        """Safely parses the semantic version str. The epoch and local
        will be removed.

        :param ver: version str. Best to preprocess this using sanitize_tag
        :type ver: str
        :param local:

           Default None. local format ``+g[commit].d[YYYYMMDD]``.
           Format wrong if there isn't minimum one tagged version

        :type local: str | None
        :raises:

           - :py:exc:`ValueError` -- Invalid version string. git
             requires one commit. setupttools-scm requires one tagged commit

        """
        # tuple[tuple[int, int, int, str, int], int]
        t_ver, dev = get_version(ver, is_use_final=self.is_use_final)
        # releaselevel is long form: post, candidate, alpha, beta
        # releaselevel could be in short form: a, b, rc
        major, minor, micro, releaselevel, serial = t_ver
        self._major = major
        self._minor = minor
        self._micro = micro
        self._releaselevel = releaselevel
        self._serial = serial
        self._dev = dev
        if local is None:  # pragma: no cover
            self._local = None
        else:
            is_unsupported = not isinstance(local, str)
            is_empty_str = isinstance(local, str) and len(local.strip()) == 0
            if is_unsupported or is_empty_str:  # pragma: no cover
                self._local = None
            else:
                self._local = local

        self._release = (major, minor, micro)

    def version_xyz(self):
        """Get xyz version. Call parse_ver first

        :returns: xyz version str. Only None if parse_ver not called beforehand
        :rtype: str | None
        """
        is_major = (
            hasattr(self, "_major")
            and self._major is not None
            and isinstance(self._major, int)
            and self._major >= 0
        )
        is_minor = (
            hasattr(self, "_minor")
            and self._minor is not None
            and isinstance(self._minor, int)
            and self._minor >= 0
        )
        is_micro = (
            hasattr(self, "_micro")
            and self._micro is not None
            and isinstance(self._micro, int)
            and self._micro >= 0
        )

        if is_major and is_minor and is_micro:
            ret = f"{self._major}.{self._minor}.{self._micro}"
        else:
            ret = None

        return ret

    def anchor(self):
        """Full semantic version for display. xyz separated by hyphens

        :returns: None when parse_var has not been called otherwise version for display
        :rtype: str | None
        """
        xyz = self.version_xyz()
        if xyz is None:
            # Only None if parse_ver not called beforehand
            ret = None
        else:
            ret = xyz.replace(".", "-")

            # a, b, rc, post
            rel = self.releaselevel_abbr

            is_rel = rel is not None and isinstance(rel, str) and len(rel) != 0
            if is_rel and rel != "final":
                ret += f"{rel}{self.serial}"

        return ret

    def readthedocs_url(self, package_name, is_latest=False):
        """Get readthedocs.io URL. Call parse_ver first

        :param package_name:

           Differences from app name, contains hyphens rather than underscores

        :param is_latest:

           Default False. If True get latest url otherwise get version specific url

        :type is_latest: bool | None
        :returns:

           url to readthedocs.io for a semantic version of the docs, not
           necessarily the latest

        :rtype: str
        """
        # app-name --> package_name
        if "_" in package_name:
            package_name = package_name.replace("_", "-")
        else:  # pragma: no cover
            pass

        if is_latest is None or not isinstance(is_latest, bool):
            bol_is_latest = False
        else:
            bol_is_latest = is_latest

        if bol_is_latest is True:
            ver_ = "latest"
        else:
            ver_ = self.version_xyz()
            if ver_ is None:
                ver_ = "latest"
            else:  # pragma: no cover
                pass

        ret = f"https://{package_name}.readthedocs.io/en/{ver_}"

        return ret

    def version_clean(self, kind):
        """Gets cleaned version str from either git or version_file.

        - "current" or "now"
          Gets current version thru setuptools-scm

        - "tag"
          Gets version from version file. Greatly depends on the
          ``pyproject.toml`` being valid and well configured.

          If ``pyproject.toml`` doesn't exist or malformed will raise a
          :py:exc:`AssertionError`. For any other reason, will issue a
          warning and return None.

          Prefer to fallback to the current version rather than not be
          able to return a version str at all. Convert the
          warning --> strerr, so have a chance to correct the problem
          without sacrificing usability

        - a version str
          Uses that version str. Use this to create prerelease, post
          release, and tagged versions

        :param kind: a known kind or a version str
        :type kind: str
        :returns: cleaned version str
        :rtype: str
        :raises:

           - :py:exc:`AssertionError` -- ``pyproject.toml`` missing or malformed

           - :py:exc:`ValueError` -- Explicit version str invalid

        """
        cls = type(self)

        # sanitize again just in case
        # returns: current, tag, or version str
        kind_: str = cls.sanitize_kind(kind)

        if kind_ not in cls.KINDS:
            # Version str. Will become next version
            git_ver = kind_
            try:
                clean_ver, local = sanitize_tag(git_ver)
            except ValueError:
                self._version = None
                raise
        else:
            # Don't sanitize. Affects src/[project name]/_version.py
            # Automagically written by setuptools-scm. Version -- git
            if kind_ == cls.CURRENT_ALIAS_DEFAULT:
                """Most likely development version
                No git init, current version --> "0.0.1"
                """
                func = partial(_current_version, path=self.path_cwd)
                git_ver = func()
            else:
                """last tagged version. May throw AttributeError if
                has pyproject.toml configuration issues"""
                func = partial(
                    _tag_version,
                    path=self.path_cwd,
                )
                git_ver = func()
                # Check -- version file semantic version str
                if git_ver is not None:
                    try:
                        sanitize_tag(git_ver)
                    except ValueError:
                        # invalid semantic version str in version file
                        git_ver = None
                else:  # pragma: no cover
                    pass

                if git_ver is None:
                    """No tagged version. Fallback to current version.
                    No git init, current version --> "0.0.1"
                    """
                    func = partial(_current_version, path=self.path_cwd)
                    git_ver = func()
                else:  # pragma: no cover
                    pass

            clean_ver, local = sanitize_tag(git_ver)

        self._version = clean_ver

        # w/o final
        return clean_ver

    @property
    def __version__(self):
        """Cleaned version available after call to version_clean

        :returns: Cleaned version or None if version_clean has not been called
        :rtype: str | None
        """
        return self._version



import types
from pathlib import Path
from typing import Any

__all__ = (
    "SemVersion",
    "sanitize_tag",
    "get_version",
)

_map_release: types.MappingProxyType[str, str]

def _is_package(package_name: str) -> bool: ...
def _is_setuptools_scm() -> bool: ...
def _path_or_cwd(val: Any | None) -> Path: ...
def get_package_name(path: Path) -> str | None: ...
def _scm_key(prog_name: str) -> str: ...
def _current_tag(path: Path | None = None) -> str | None: ...
def _strip_epoch(ver: str) -> tuple[str | None, str]: ...
def _strip_local(ver: str) -> tuple[str | None, str]: ...
def _remove_v(ver: str) -> str: ...
def sanitize_tag(ver: str) -> tuple[str, str | None]: ...
def get_version(
    ver: str,
    is_use_final: bool = False,
) -> tuple[tuple[int, int, int, str, int], int | None]: ...
def _current_version(path: Path | None = None) -> str | None: ...
def _tag_version(
    next_version: str | None = "",
    path: Path | None = None,
) -> str | None: ...
def _arbritary_version(
    next_version: str,
    path: Path | None = None,
) -> str | None: ...
def _get_app_name(path: Path | None = None) -> str | None: ...

class SemVersion:
    CURRENT_ALIAS_DEFAULT: str = "current"
    CURRENT_ALIASES: tuple[str, str] = ...
    KINDS: tuple[str, str, str] = ...

    def __init__(
        self,
        path: Path | None = None,
        is_use_final: Any | None = False,
    ) -> None: ...
    @classmethod
    def sanitize_kind(cls, kind: str | None = None) -> str: ...
    @property
    def path_cwd(self) -> Path: ...
    @path_cwd.setter
    def path_cwd(self, val: Any) -> None: ...
    @property
    def is_use_final(self) -> bool: ...
    @is_use_final.setter
    def is_use_final(self, val: Any) -> None: ...
    @property
    def major(self) -> int | None: ...
    @property
    def minor(self) -> int | None: ...
    @property
    def micro(self) -> int | None: ...
    @property
    def releaselevel(self) -> str | None: ...
    @releaselevel.setter
    def releaselevel(self, val: str) -> None: ...
    @property
    def releaselevel_abbr(self) -> str: ...
    @property
    def serial(self) -> int | None: ...
    @property
    def dev(self) -> int | None: ...
    @property
    def release(self) -> tuple[int, int, int]: ...
    @staticmethod
    def as_tuple(version_str: str) -> tuple[int | str, ...]: ...
    def parse_ver(self, ver: str, local: str | None = None) -> None: ...
    def version_xyz(self) -> str | None: ...
    def anchor(self) -> str | None: ...
    def readthedocs_url(
        self,
        package_name: str,
        is_latest: bool | None = False,
    ) -> str: ...
    def version_clean(
        self,
        kind: str,
    ) -> str | None: ...
    @property
    def __version__(self) -> str | None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Provides primitives to use hooks to extend behaviour

Built-in plugins are lazy loaded. Args must be passed in using a workaround

config_settings dict would normally contain cli args passed thru by

.. code-block:: text

   python -m build -C--[option name]="[value]"

However setuptools has not implemented support for this. The build
process hooks are each ran in a subprocess. :code:`sys.argv` is not
modified to pass thru the config settings. As a temporary workaround,
**that doesn't add functionality**, a temporary config file is created
and an environment variable containing path to the temporary config file
is passed to :code:`python -m build`

.. code-block:: shell

   export DS_CONFIG_SETTINGS=/tmp/setuptools-build.toml
   cat <<-EOF > "$DS_CONFIG_SETTINGS"
   [project]
   name = "whatever"
   version = "99.99.99a1.dev6"

   [tool.config-settings]
   kind="0.0.1"
   set-lock="0"
   EOF
   python -m build --no-isolation

.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("_PluggyPlugin", "_create_hook_manager", "hook_impl")

   This sub-package exports

.. seealso::

   https://docs.kedro.org/en/latest/hooks/introduction.html
   https://github.com/kedro-org/kedro/blob/ab2e4798c2670639692b2097cae7604045434b18/kedro/framework/project/__init__.py#L97

"""

from .manager import _create_hook_manager
from .markers import hook_impl

__all__ = (
    "_create_hook_manager",
    "hook_impl",
)



"""
.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_wrap_version_keyword.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_wrap_version_keyword.py

.. seealso::

   https://github.com/pypa/setuptools/blob/main/setuptools/tests/test_dist.py
   https://setuptools.pypa.io/en/latest/userguide/package_discovery.html#package-discovery-and-namespace-packages
   https://github.com/pypa/setuptools_scm/blob/main/testing/test_integration.py

"""

import logging
import logging.config
from collections.abc import Mapping
from pathlib import Path
from unittest.mock import (
    MagicMock,
    patch,
)

import pytest  # noqa: F401
from setuptools.dist import Distribution
from setuptools_scm._config import Configuration

from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.monkey.patch_pyproject_reading import ReadPyprojectStrict
from drain_swamp.monkey.wrap_version_keyword import version_keyword


@pytest.fixture
def version_file():
    """Ensure the temporary version file is removed"""
    path_root_dir = Path(__file__).parent.parent
    path_f = path_root_dir.joinpath("src", "_version.py")
    path_f.touch()
    yield path_f
    path_root_dir.joinpath("src", "_version.py").unlink()


def test_version_keyword(caplog, has_logging_occurred, version_file):
    """setuptools-scm would normally needs a setup.py to be executed"""
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    path_root_dir = Path(__file__).parent.parent
    path_version_file_relpath = str(version_file.relative_to(path_root_dir))
    root_abspath = str(path_root_dir)
    package_dir = {"": "src"}
    attrs = {
        "src_root": root_abspath,
        "package_dir": package_dir,
    }
    key = "use_scm_version"

    # can be dict or True
    value = {"important": "stuff"}
    assert isinstance(value, Mapping)

    # from importlib.metadata import distribution
    # dist_0 = distribution(g_app_name)
    # pkg_name = dist_0.name
    # ver_pkg = dist_0.version
    # logger.info(f"dist.name: {pkg_name}")
    # logger.info(f"dist.version: {dist_0.version}")

    dist = Distribution(attrs)
    dist.set_defaults()

    relative_to = "pyproject.toml"
    data = {"root": root_abspath, "write_to": path_version_file_relpath}
    with (
        # setuptools_scm._integration.setuptools._assign_version
        patch(
            "setuptools_scm._integration.setuptools._config._read_pyproject",
            new_callable=MagicMock(wraps=ReadPyprojectStrict),
        ),
        # setuptools_scm._config.Configuration.from_file
        patch(
            "setuptools_scm._config._read_pyproject",
            new_callable=MagicMock(wraps=ReadPyprojectStrict),
        ),
        patch(
            "setuptools_scm._config.Configuration.from_data",
            return_value=Configuration.from_data(relative_to, data),
        ),
    ):
        # getter; not a setter. So src/_version.py not written to
        version_keyword(dist, key, value)

    assert isinstance(dist.metadata.version, str)
    assert len(dist.metadata.version) != 0
    # assert has_logging_occurred(caplog)



from collections.abc import Callable
from typing import (
    TypeVar,
    overload,
)

_F = TypeVar("_F", bound=Callable[..., object])

# From types-pluggy
class HookimplMarker:
    project_name: str
    def __init__(self, project_name: str) -> None: ...
    @overload
    def __call__(
        self,
        function: _F,
        hookwrapper: bool = False,
        optionalhook: bool = False,
        tryfirst: bool = False,
        trylast: bool = False,
        specname: str | None = None,
        wrapper: bool = False,
    ) -> _F: ...
    @overload
    def __call__(
        self,
        function: None = None,
        hookwrapper: bool = False,
        optionalhook: bool = False,
        tryfirst: bool = False,
        trylast: bool = False,
        specname: str | None = None,
        wrapper: bool = False,
    ) -> Callable[[_F], _F]: ...

hook_impl: HookimplMarker  # noqa: E305



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Keep constants out of package module, ``__init__.py``

.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("_PLUGIN_HOOKS", "DOTTED_PATH_SPECS", "HOOK_NAMESPACE")

   Module exports

.. py:data:: _PLUGIN_HOOKS
   :type: str
   :value: "ds.hooks"

   Entrypoint name, not the dotted path:fcn. Entrypoint underwhich to specify additional hooks

.. py:data:: HOOK_NAMESPACE
   :type: str
   :value: "ds"

   .. seealso::

      namespace
      `naming conversion <https://pluggy.readthedocs.io/en/latest/#a-complete-example>`_


.. py:data:: DOTTED_PATH_SPECS
   :type: str
   :value: "drain_swamp.monkey.hooks.specs"

   Dotted path to pluggy plugin specs

"""

__all__ = (
    "_PLUGIN_HOOKS",
    "DOTTED_PATH_SPECS",
    "HOOK_NAMESPACE",
)

_PLUGIN_HOOKS = "ds.hooks"
DOTTED_PATH_SPECS = "drain_swamp.monkey.hooks.specs"
HOOK_NAMESPACE = "ds"



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. py:data:: __all__
   :type: tuple[str, str, str, str]
   :value: ("PyProjectTOMLParseError", "BackendNotSupportedError", \
    "PyProjectTOMLReadError", "MissingRequirementsFoldersFiles")

   Package level exports

.. seealso::

   `pep366 <https://peps.python.org/pep-0366/>`_

   `__package__ usage <https://stackoverflow.com/a/48833828>`_

   `import package <https://stackoverflow.com/a/6655098>`_

   `removal of __package__ <https://github.com/python/cpython/pull/97879>`_

"""

__package__ = __name__

from .constants_maybe import __version_app as __version__  # noqa: F401
from .exceptions import (
    BackendNotSupportedError,
    MissingRequirementsFoldersFiles,
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)

__all__ = (
    "PyProjectTOMLParseError",
    "BackendNotSupportedError",
    "PyProjectTOMLReadError",
    "MissingRequirementsFoldersFiles",
)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

``igor.py`` deals with updating:

- Sphinx ``doc?/conf.py``

- ``NOTICE.txt``

- ``CHANGES.rst``

Already have semantic version handler. For ``NOTICE.txt`` and
``CHANGES.rst``, only lack the authors left name. Which is usually first
name, in Japanese, it's the ``[last][first]`` name.

.. py:data:: __all__
   :type: tuple[str]
   :value: ("PackageMetadata",)

   Module exports

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

.. py:data:: AUTHOR_NAME_FALLBACK
   :type: str
   :value: "Anonymous"

   Fallback when unable to get author name, from either package data or pyproject.toml


"""

import copy
import logging
from email.utils import parseaddr
from functools import cache
from importlib.metadata import (
    PackageNotFoundError,
    metadata,
)

from .check_type import is_ok
from .constants import g_app_name
from .parser_in import TomlParser

_logger = logging.getLogger(f"{g_app_name}.package_metadata")

__package__ = "drain_swamp"
__all__ = ("PackageMetadata",)
AUTHOR_NAME_FALLBACK = "Anonymous"


def is_package_installed(app_name):
    try:
        metadata(app_name)
    except PackageNotFoundError:
        ret = False
    else:
        ret = True

    return ret


@cache
def get_author_and_email(app_name):
    """Get author name and email from package metadata.

    :param app_name:

       Contains only underscore, not hyphens. Cuz used by
       :py:func:`functools.cache` as a key. So that key must be unique
       across multiple calls

    :type app_name: str
    :returns: Author name, fallback is "Anonymous" and email fallback is None
    :rtype: tuple[str, str | None]
    """
    try:
        package_meta = metadata(app_name)
    except PackageNotFoundError:
        """Package not installed. Fallback to attempting to get from
        ``pyproject.toml``"""
        msg_exc = (
            f"Package, {app_name}, not installed. Could not get author name "
            "from Python package metadata. Default --> Anonymous"
        )
        _logger.warning(msg_exc)
        author_name = AUTHOR_NAME_FALLBACK
        author_email = None
    else:
        rfc2822_addr = package_meta["Author-email"]
        author_name, author_email = parseaddr(rfc2822_addr)

    return (author_name, author_email)


class PackageMetadata:
    """Access author details from either package metadata or cache

    .. py:attribute:: __slots__
       :type: tuple[str, str, str]
       :value: ("_app_name", "_full_name", "_email")

    :vartype app_name:

       Key used to get the author details either package metadata or cache

    :ivar: typing.Any
    :vartype path:

       Default None. If the package has yet to be installed, fallback
       to ``pyproject.toml``

    :ivar: path: pathlib.Path | None
    """

    __slots__ = ("_app_name", "_full_name", "_email", "_d_pyproject_toml")

    def __init__(self, app_name, path=None):
        super().__init__()
        self.app_name = app_name

        if is_package_installed(self.app_name):
            self._full_name, self._email = get_author_and_email(self.app_name)
        else:
            # Looks alot like :py:func:`drain_swamp.pep518_read.find_pyproject_toml`
            tp = TomlParser(path)
            d_pyproject_toml = tp.d_pyproject_toml
            if d_pyproject_toml is None:
                self._full_name = AUTHOR_NAME_FALLBACK
                self._email = None
                return

            self._d_pyproject_toml = copy.deepcopy(d_pyproject_toml)

            authors = d_pyproject_toml.get("project", {}).get("authors", [])
            if isinstance(authors, list) and len(authors) >= 1:
                d_author = authors[0]
                self._full_name = d_author["name"]
                self._email = d_author["email"]
            else:
                msg_exc = (
                    "Package not installed and pyproject.toml project.authors "
                    "does not exist or is invalid"
                )
                _logger.warning(msg_exc)
                self._full_name = AUTHOR_NAME_FALLBACK
                self._email = None

    @property
    def app_name(self):
        """Alphanumeric + underscores. Used as a key

        :returns: app name
        :rtype: str
        """
        return self._app_name

    @app_name.setter
    def app_name(self, val):
        """Setter for app name. Which is the key used to get the author
        details from cache.

        :param val: Either package name (with hyphens) or app name (with underscores)
        :type val: typing.Any
        """
        if is_ok(val):
            self._app_name = val.replace("-", "_")
        else:
            msg_exc = f"Expecting non-empty str got type {type(val)}"
            raise ValueError(msg_exc)

    @property
    def full_name(self):
        """Author name from package metadata. So need to supply in ``pyproject.toml``

        For example,

        .. code-block:: text

           [project]
           authors = [  # Contact by mastodon please
               {name = "Dave Faulkmore", email = "faulkmore@protonmail.com"},
           ]

        :returns: full author name
        :rtype: str
        """
        return self._full_name

    @property
    def left_name(self):
        """Why not first name? Well in Japanese, name is written last
        then first name. In which case return last name.

        Intended to be used with a regex to parse a Copyright line

        .. code-block:: text

           regex_pattern = r"Copyright {0}.*? {1}".format(
               copyright_start_year,
               author_first_name,
           )

        :returns: Left name which is usually first name, except for Japanese names.
        :rtype: str

        """
        name = self._full_name
        names = name.split()
        ret = names[0]
        return ret

    @property
    def email(self):
        """Author email address as found in package metadata

        :returns:

           Author email address. To force author to comply and learn
           what's an abacus. Probably useless for contacting the author

           Companies send rejection emails exactly one to two days after
           applying. Automated cron job. The companies are brand
           marketing; there is no job, recruiter, or HR manager. Just
           an unimaginative cron job

        :rtype: str | None
        """
        return self._email

    @property
    def d_pyproject_toml(self):
        """So don't have to get ``pyproject.toml`` repeatedly

        :returns: pyproject.toml as a dict otherwise None
        :rtype: dict[str, typing.Any] | None
        """
        return self._d_pyproject_toml



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Unittest for module, drain_swamp.snip

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_snip.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_snip.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

"""

import contextlib
import io
import logging
import logging.config
from pathlib import Path

import pytest

from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)
from drain_swamp.snip import (
    ReplaceResult,
    Snip,
    check_matching_tag_count,
    check_not_nested_or_out_of_order,
)

testdata_test_snip_harden = [
    (
        "with_id_key_no_snippet.txt",
        "asdf",
        "abc abc abc",
        "zzzzzzzzzzzzz\nzzzzzzzzzz\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\nzzzzzzzzzz\nzzzzzzzzzzzzz\n",
        ReplaceResult.NO_MATCH,
        False,
    ),
    (
        "with_id_key_no_key.txt",
        "asdf",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.NO_MATCH,
        False,
    ),
    (
        "with_id_no_key_key.txt",
        "",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable george\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable george\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        False,
    ),
    (
        "with_id_no_change.txt",
        "asdf",
        "blah blah blah",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.NO_CHANGE,
        False,
    ),
    (
        "with_id.txt",
        "asdf",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        False,
    ),
    (
        "without_id_have_id_no_match.txt",
        "george",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\nzzzzzzzzzzzzz\n# @@@ editable george\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\nzzzzzzzzzzzzz\n# @@@ editable george\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        True,
    ),
    (
        "with_id_snippet_empty.txt",
        "asdf",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\n\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        False,
    ),
    (
        "without_id_empty_str.txt",
        "",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        True,
    ),
    (
        "without_id_none.txt",
        None,
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        True,
    ),
    (
        "without_id_not_str.txt",
        1.12345,
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        True,
    ),
    (
        "without_id_excess_whitespace.txt",
        "    ",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable\nabc abc abc\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.REPLACED,
        True,
    ),
    (
        "with_id_match_2nd.txt",
        "george",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nabc abc abc\n# @@@ end\nzzzzzzz\n",
        ReplaceResult.REPLACED,
        False,
    ),
    (
        "with_id_match_1st.txt",
        "asdf",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nabc abc abc\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n",
        ReplaceResult.REPLACED,
        False,
    ),
    (
        "with_id_match_1st_and_3rd.txt",
        "asdf",
        "abc abc abc",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n# @@@ editable asdf\nlol lol lol\n# @@@ end\noh ok then",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nabc abc abc\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n# @@@ editable asdf\nabc abc abc\n# @@@ end\noh ok then",
        ReplaceResult.REPLACED,
        False,
    ),
]
ids_test_snip_harden = [
    "No snippet. Nothing to do",
    "one snip. with id. key / no key",
    "one snip. with id. no key / key",
    "one snip. with id. no change",
    "one snip. with id",
    "two snippets. one with one without id. key. invalid to mix",
    "one snip. with id. snippet empty",
    "one snip. without id empty str",
    "one snip. without id empty none",
    "one snip. without id empty not a str",
    "one snip. without id empty excess whitespace",
    "two snips. both have id. match 2nd",
    "two snips. both have id. match 1st",
    "three snips. All have id. match 1st and 3rd",
]


@pytest.mark.parametrize(
    "file_name, id_, replace_text, file_contents, expected, expected_result_status, is_quiet",
    testdata_test_snip_harden,
    ids=ids_test_snip_harden,
)
def test_snip_harden(
    file_name,
    id_,
    replace_text,
    file_contents,
    expected,
    expected_result_status,
    is_quiet,
    tmp_path,
    caplog,
    file_regression,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_snip_harden" tests
    # py39+ Cannot have LOGGING.loggers.root
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    path_fname = tmp_path.joinpath(file_name)
    path_fnames = (
        path_fname,
        str(path_fname),
    )
    for mixed_f in path_fnames:
        Path(mixed_f).touch()
        Path(mixed_f).write_text("")  # reset file contents

        # empty file --> ValueError
        er_bad_val = Snip(mixed_f, is_quiet=is_quiet)
        with pytest.raises(ValueError):
            er_bad_val.get_file()

        # files/folders --> FileNotFoundError --> False
        invalids = (
            str(tmp_path),  # a folder
            Path("/proc/tmp/george/bob/ted/todd/dan.txt"),  # nonexistent file
            Path("__init__.py"),  # relative file (not scary normally empty)
        )
        for invalid in invalids:
            er_bad = Snip(invalid, is_quiet=is_quiet)
            # Fails at this point, not on class construction
            is_success = er_bad.replace(replace_text, id_=id_)
            assert is_success is None or ReplaceResult.VALIDATE_FAIL == is_success

        Path(mixed_f).write_text(file_contents)
        er = Snip(mixed_f, is_quiet=is_quiet)

        # pass in non-str
        invalids = (
            None,
            1.1234,
            1,
        )
        for invalid in invalids:
            with pytest.raises(TypeError):
                er.replace(invalid, id_=id_)

        # normal call
        is_success = er.replace(replace_text, id_=id_)
        if is_success is None or is_success == ReplaceResult.VALIDATE_FAIL:
            # issue with file or one or more validation checks failed
            pass
        else:
            assert is_success == expected_result_status
            actual = er.get_file()
            assert actual == expected

            assert has_logging_occurred(caplog)

            # https://pytest-regressions.readthedocs.io/en/latest/api.html#pytest_regressions.file_regression.FileRegressionFixture.check
            file_regression.check(expected, extension=".txt", binary=False)


def test_snip_properties():
    # pytest --showlocals --log-level INFO -k "test_properties" tests
    is_quiet_invalids = (
        None,
        "Hello World!",
    )
    path_invalids = (
        None,
        1.12345,
        1,
    )
    for p_invalid in path_invalids:
        for q_invalid in is_quiet_invalids:
            with pytest.raises(TypeError):
                Snip(p_invalid, is_quiet=q_invalid)


@pytest.mark.parametrize(
    "file_name, id_, replace_text, file_contents, expected, expected_result_status, is_quiet",
    testdata_test_snip_harden,
    ids=ids_test_snip_harden,
)
def test_checks_normal_usage(
    file_name,
    id_,
    replace_text,
    file_contents,
    expected,
    expected_result_status,
    is_quiet,
    caplog,
):
    # pytest --showlocals --log-level INFO -k "test_checks_normal_usage" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    assert check_matching_tag_count(
        file_contents,
        Snip.TOKEN_START,
        Snip.TOKEN_END,
    )

    # check_matching_tag_count must occur before this check
    #    arg validation -- invalids
    invalids = (
        None,
        "",
        "    ",
        1.12345,
    )
    test_data = []
    a_ = file_contents
    b_ = Snip.TOKEN_START
    c_ = Snip.TOKEN_END
    for invalid_a in invalids:
        test_data.append((invalid_a, b_, c_))
    for invalid_b in invalids:
        test_data.append((a_, invalid_b, c_))
    for invalid_c in invalids:
        test_data.append((a_, b_, invalid_c))
    for t_args in test_data:
        assert check_not_nested_or_out_of_order(*t_args) is False

    #    normal usage
    assert check_not_nested_or_out_of_order(
        file_contents,
        Snip.TOKEN_START,
        Snip.TOKEN_END,
    )


@pytest.mark.parametrize(
    "file_name, id_, replace_text, file_contents, expected, expected_result_status, is_quiet",
    testdata_test_snip_harden,
    ids=ids_test_snip_harden,
)
def test_checks_bad_input(
    file_name,
    id_,
    replace_text,
    file_contents,
    expected,
    expected_result_status,
    is_quiet,
    caplog,
):
    # pytest --showlocals --log-level INFO -k "test_checks_bad_input" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    # prepare test data
    invalids = (
        None,
        "",
        "    ",
        1.12345,
    )
    test_data_checks_bad_input = []
    a_ = file_contents
    b_ = Snip.TOKEN_START
    c_ = Snip.TOKEN_END
    for invalid_a in invalids:
        test_data_checks_bad_input.append((invalid_a, b_, c_))
    for invalid_b in invalids:
        test_data_checks_bad_input.append((a_, invalid_b, c_))
    for invalid_c in invalids:
        test_data_checks_bad_input.append((a_, b_, invalid_c))

    # check_matching_tag_count
    for t_args in test_data_checks_bad_input:
        assert check_matching_tag_count(*t_args) is False

    # check_not_nested_or_out_of_order
    for t_args in test_data_checks_bad_input:
        assert check_not_nested_or_out_of_order(*t_args) is False


SNIPS_BAD = list(Path(__file__).parent.joinpath("_bad_snips").glob("*.txt"))


@pytest.mark.parametrize(
    "path",
    SNIPS_BAD,
    ids=[path.name.rsplit(".", 1)[0] for path in SNIPS_BAD],
)
def test_check_snips_bad(path):
    """Snips gone wrong."""
    # pytest --showlocals --log-level INFO -k "test_check_snips_bad" tests

    file_contents = path.read_text()
    invalids = [
        (file_contents, Snip.TOKEN_START, Snip.TOKEN_END),
    ]

    for t_args in invalids:
        # check_matching_tag_count ... pass
        assert check_matching_tag_count(*t_args)

        # check_not_nested_or_out_of_order ... fail
        assert not check_not_nested_or_out_of_order(*t_args)

    # bad files should fail validation
    snip = Snip(path)
    assert snip.validate() is False


@pytest.mark.parametrize(
    "file_name, id_, replace_text, file_contents, expected, expected_result_status, is_quiet",
    testdata_test_snip_harden,
    ids=ids_test_snip_harden,
)
def test_snip_validate(
    file_name,
    id_,
    replace_text,
    file_contents,
    expected,
    expected_result_status,
    is_quiet,
    caplog,
    tmp_path,
):
    # pytest --showlocals --log-level INFO -k "test_snip_validate" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    path_fname = tmp_path.joinpath(file_name)
    path_fnames = (
        path_fname,
        str(path_fname),
    )
    for mixed_f in path_fnames:
        Path(mixed_f).touch()
        Path(mixed_f).write_text("")  # reset file contents

        # empty file --> ValueError --> False
        snip_empty = Snip(mixed_f, is_quiet=is_quiet)
        assert snip_empty.validate() is False

        Path(mixed_f).write_text(file_contents)

        # good
        snip = Snip(mixed_f, is_quiet=True)
        is_valid = snip.validate()
        assert is_valid is True

        # invalid files/folders --> FileNotFoundError --> False
        invalids = (
            str(tmp_path),  # a folder
            Path("/proc/tmp/george/bob/ted/todd/dan.txt"),  # nonexistent file
            Path("__init__.py"),  # relative file (not scary normally empty)
        )
        for invalid in invalids:
            snip_bad = Snip(invalid, is_quiet=is_quiet)
            # Fails at this point, not on class construction
            is_valid = snip_bad.validate()
            assert is_valid is False


testdata_snip_contents = [
    (
        "with_id_key_no_snippet.txt",
        "asdf",
        "zzzzzzzzzzzzz\nzzzzzzzzzz\nzzzzzzzzzzzzz\n",
        ReplaceResult.NO_MATCH,
        False,
    ),
    (
        "with_id_key_no_key.txt",
        "asdf",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        ReplaceResult.NO_MATCH,
        False,
    ),
    (
        "with_id_no_key_key.txt",
        "",
        "zzzzzzzzzzzzz\n# @@@ editable george\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "blah blah blah",
        False,
    ),
    (
        "with_id.txt",
        "asdf",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "blah blah blah",
        False,
    ),
    (
        "with_id_snippet_empty.txt",
        "asdf",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\n\n# @@@ end\nzzzzzzzzzzzzz\n",
        "",
        False,
    ),
    (
        "without_id_empty_str.txt",
        "",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "blah blah blah",
        True,
    ),
    (
        "without_id_none.txt",
        None,
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "blah blah blah",
        True,
    ),
    (
        "without_id_not_str.txt",
        1.12345,
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "blah blah blah",
        True,
    ),
    (
        "without_id_excess_whitespace.txt",
        "    ",
        "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzzz\n",
        "blah blah blah",
        True,
    ),
    (
        "with_id_match_2nd.txt",
        "george",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n",
        "hey there ted",
        False,
    ),
    (
        "with_id_match_1st.txt",
        "asdf",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n",
        "blah blah blah",
        False,
    ),
    (
        "with_id_match_1st_and_3rd.txt",
        "asdf",
        "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzz\n# @@@ editable george\nhey there ted\n# @@@ end\nzzzzzzz\n# @@@ editable asdf\nlol lol lol\n# @@@ end\noh ok then",
        "blah blah blah",
        False,
    ),
]
ids_snip_contents = [
    "No snippets. Nothing to do",
    "one snip. with id. No key",
    "one snip. with id. No key. Infers key",
    "one snip. with id",
    "one snip. with id. snippet empty",
    "one snip. without id empty str. Infers key",
    "one snip. without id empty none/ Infers key",
    "one snip. without id empty not a str. Infers key",
    "one snip. without id empty excess whitespace. Infers key",
    "two snips. both have id. match 2nd",
    "two snips. both have id. match 1st",
    "three snips. All have id. match 1st and 3rd",
]


@pytest.mark.parametrize(
    "file_name, id_, file_contents, expected, is_quiet",
    testdata_snip_contents,
    ids=ids_snip_contents,
)
def test_snip_contents(
    file_name,
    id_,
    file_contents,
    expected,
    is_quiet,
    tmp_path,
    prepare_folders_files,
    caplog,
    has_logging_occurred,
):
    # pytest --showlocals --log-level INFO -k "test_snip_contents" tests
    # pytest --showlocals --log-level INFO tests/test_snip.py::test_snip_contents["No snippets. Nothing to do"]
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    path_abs = tmp_path / file_name
    snip = Snip(path_abs, is_quiet=is_quiet)

    # No preparation -- snip.contents
    t_actual = snip.contents(id_=id_)
    assert t_actual == ReplaceResult.VALIDATE_FAIL

    # No preparation -- snip.print
    with contextlib.redirect_stderr(io.StringIO()):
        snippets = snip.print()
    assert snippets == ReplaceResult.VALIDATE_FAIL

    # prepare
    seq_create_these = (file_name,)
    prepare_folders_files(seq_create_these, tmp_path)
    path_abs.write_text(file_contents)

    # act -- snip.print
    #    same return value as snip.snippet
    with contextlib.redirect_stderr(io.StringIO()) as f_1:
        snippets = snip.print()
    msg = f_1.getvalue()

    if isinstance(snippets, ReplaceResult):
        assert snippets == ReplaceResult.NO_MATCH
        # logger.info(f"No snippets: {snippets}\n{msg}")
        pass
    else:
        #    Block of human readable beautiful prose was printed
        assert len(msg) != 0

    # act -- snip.contents
    t_actual = snip.contents(id_=id_)

    assert has_logging_occurred(caplog)

    if isinstance(t_actual, ReplaceResult):
        assert t_actual == ReplaceResult.NO_MATCH
    else:
        # id_ might have been inferred
        assert t_actual[0] == expected
